#!/usr/bin/env python2.7
# This script builds the distribution packages platform-independently.
# No parameters needed; config is auto-detected.

import sys
import os
import platform
import shutil
import time
import glob
import build_version
import build_number
import builder.utils

# Configuration.
LAUNCH_DIR    = os.path.abspath(os.getcwd())
DOOMSDAY_DIR  = os.path.abspath(os.path.join(os.getcwd(), '..', 'doomsday'))
SNOWBERRY_DIR = os.path.abspath(os.path.join(LAUNCH_DIR, '..', 'snowberry'))
WORK_DIR      = os.path.join(LAUNCH_DIR, 'work')
OUTPUT_DIR    = os.path.abspath(os.path.join(os.getcwd(), 'releases'))
DOOMSDAY_VERSION_FULL       = "0.0.0-Name"
DOOMSDAY_VERSION_FULL_PLAIN = "0.0.0"
DOOMSDAY_VERSION_MAJOR      = 0
DOOMSDAY_VERSION_MINOR      = 0
DOOMSDAY_VERSION_REVISION   = 0
DOOMSDAY_RELEASE_TYPE       = "Unstable"
DOOMSDAY_BUILD_NUMBER       = build_number.todays_build()
DOOMSDAY_BUILD              = 'build' + DOOMSDAY_BUILD_NUMBER

TIMESTAMP = time.strftime('%y-%m-%d')
now = time.localtime()


def exit_with_error():
    os.chdir(LAUNCH_DIR)
    sys.exit(1)


def mkdir(n):
    try:
        os.mkdir(n)
    except OSError:
        print 'Directory', n, 'already exists.'


def remkdir(n):
    builder.utils.remkdir(n)


def remove(n):
    try:
        os.remove(n)
    except OSError:
        print 'Cannot remove', n


def copytree(s, d):
    try:
        shutil.copytree(s, d)
    except Exception, x:
        print x
        print 'Cannot copy', s, 'to', d


def duptree(s, d):
    os.system('cp -fRp "%s" "%s"' % (s, d))


def find_version():
    build_version.find_version()

    global DOOMSDAY_VERSION_FULL
    global DOOMSDAY_VERSION_FULL_PLAIN
    global DOOMSDAY_VERSION_MAJOR
    global DOOMSDAY_VERSION_MINOR
    global DOOMSDAY_VERSION_REVISION
    global DOOMSDAY_RELEASE_TYPE

    DOOMSDAY_RELEASE_TYPE = build_version.DOOMSDAY_RELEASE_TYPE
    DOOMSDAY_VERSION_FULL_PLAIN = build_version.DOOMSDAY_VERSION_FULL_PLAIN
    DOOMSDAY_VERSION_FULL = build_version.DOOMSDAY_VERSION_FULL
    DOOMSDAY_VERSION_MAJOR = build_version.DOOMSDAY_VERSION_MAJOR
    DOOMSDAY_VERSION_MINOR = build_version.DOOMSDAY_VERSION_MINOR
    DOOMSDAY_VERSION_REVISION = build_version.DOOMSDAY_VERSION_REVISION
    
    print 'Build:', DOOMSDAY_BUILD, 'on', TIMESTAMP
    print 'Version:', DOOMSDAY_VERSION_FULL_PLAIN, DOOMSDAY_RELEASE_TYPE


def prepare_work_dir():
    remkdir(WORK_DIR)
    print "Work directory prepared."


def mac_os_version():
    """Determines the Mac OS version."""
    return builder.utils.mac_os_version()
    
    
def mac_os_8_or_later():
    return mac_os_version() in ['10.8', '10.9', '10.10', '10.11']


def mac_target_ext():
    if mac_os_8_or_later(): return '.dmg'
    if mac_os_version() == '10.6': return '_mac10_6.dmg'
    return '_32bit.dmg'
    
    
def mac_osx_suffix():
    if mac_os_8_or_later(): return 'macx8'
    if mac_os_version() == '10.6': return 'macx6'
    return 'macx'


def output_filename(ext='', extra=''):
    if extra != '' and not extra.endswith('_'): extra += '_'
    if DOOMSDAY_RELEASE_TYPE == "Stable":
        return 'doomsday_' + extra + DOOMSDAY_VERSION_FULL + ext
    else:
        return 'doomsday_' + extra + DOOMSDAY_VERSION_FULL + "_" + DOOMSDAY_BUILD + ext


def mac_able_to_package_snowberry():
    return mac_os_version() == '10.5'
    

def mac_package_snowberry():
    os.chdir(SNOWBERRY_DIR)
    remkdir('dist')
    remkdir('build')

    print 'Copying resources...'
    remkdir('build/addons')
    remkdir('build/conf')
    remkdir('build/graphics')
    remkdir('build/lang')
    remkdir('build/profiles')
    remkdir('build/plugins')

    for f in ['/conf/osx-appearance.conf',
              '/conf/osx-components.conf',
              '/conf/osx-doomsday.conf',
              '/conf/snowberry.conf',
              '/conf/Zcommon-doomsday.conf']:
        shutil.copy(SNOWBERRY_DIR + f, 'build/conf')

    for f in (glob.glob(SNOWBERRY_DIR + '/graphics/*.jpg') +
              glob.glob(SNOWBERRY_DIR + '/graphics/*.png') +
              glob.glob(SNOWBERRY_DIR + '/graphics/*.bmp') +
              glob.glob(SNOWBERRY_DIR + '/graphics/*.ico')):
        shutil.copy(f, 'build/graphics')

    for f in glob.glob(SNOWBERRY_DIR + '/lang/*.lang'):
        shutil.copy(f, 'build/lang')

    for f in glob.glob(SNOWBERRY_DIR + '/profiles/*.prof'):
        shutil.copy(f, 'build/profiles')

    for f in glob.glob(SNOWBERRY_DIR + '/plugins/tab*.py'):
        shutil.copy(f, 'build/plugins')

    for f in ['tab30.plugin', 'about.py', 'help.py', 'launcher.py',
              'preferences.py', 'profilelist.py', 'wizard.py']:
        src = SNOWBERRY_DIR + '/plugins/' + f
        if os.path.isdir(src):
            copytree(src, 'build/plugins/' + f)
        else:
            shutil.copy(src, 'build/plugins')

    f = file('VERSION', 'wt')
    f.write(DOOMSDAY_VERSION_FULL)
    f.close()
    builder.utils.run_python2('buildapp.py py2app')
    
    # Share it.
    duptree('dist/Doomsday Engine.app', 'shared/')


def mac_release():
    """The Mac OS X release procedure."""
    
    # Package Snowberry or acquire the shared package.
    if mac_able_to_package_snowberry():
        # Check Python dependencies.
        try:
            import wx
        except ImportError:
            raise Exception("Python: wx not found!")
        try:
            import py2app
        except ImportError:
            raise Exception("Python: py2app not found!")

        mac_package_snowberry()
        sbLoc = '/dist/Doomsday Engine.app'
    else:
        # Wait until the updated packaged SB has been shared.
        try:
            print 'This system seems unable to package Snowberry. Waiting a while'
            print 'for an updated shared Doomsday Engine.app bundle...'
            print '(press Ctrl-C to skip)'
            time.sleep(5 * 60)
        except KeyboardInterrupt:
            pass
        sbLoc = '/shared/Doomsday Engine.app'

    # First we need to make a release build.
    print "Building the release..."
    # Must work in the deng root for qmake (resource bundling apparently
    # fails otherwise).
    MAC_WORK_DIR = os.path.abspath(os.path.join(DOOMSDAY_DIR, '../macx_release_build'))
    remkdir(MAC_WORK_DIR)
    os.chdir(MAC_WORK_DIR)
        
    if os.system('PATH=`qmake-qt5 -query QT_INSTALL_BINS`:$PATH qmake -r CONFIG+=release DENG_BUILD=%s ' % (DOOMSDAY_BUILD_NUMBER) +
                 '../doomsday/doomsday.pro && PATH=`qmake-qt5 -query QT_INSTALL_BINS`:$PATH make -j2 -w'):
        raise Exception("Failed to build from source.")

    # Now we can proceed to packaging.
    target = os.path.join(OUTPUT_DIR, output_filename(mac_target_ext()))
    try:
        os.remove(target)
        print 'Removed existing target file', target
    except:
        print 'Target:', target

    # Back to the normal work dir.
    os.chdir(WORK_DIR)
    copytree(SNOWBERRY_DIR + sbLoc, 'Doomsday Engine.app')

    print 'Copying release binaries into the launcher bundle.'
    duptree(os.path.join(MAC_WORK_DIR, 'client/Doomsday.app'), 'Doomsday Engine.app/Contents/Doomsday.app')
    # Remove plugins unsuitable for general distribution.
    for omit in ['doom64', 'example']:
        builder.utils.deltree('Doomsday Engine.app/Contents/Doomsday.app/Contents/DengPlugins/%s.bundle' % omit)
    duptree(os.path.join(MAC_WORK_DIR, 'tools/shell/shell-gui/Doomsday Shell.app'), 'Doomsday Shell.app')

    print 'Correcting permissions...'
    os.system('chmod -R o-w "Doomsday Engine.app"')
    os.system('chmod -R o-w "Doomsday Shell.app"')

    if mac_os_version() != '10.6':
        print 'Packaging apps onto a disk image (unsigned)...'
        templateFile = 'appdisk.sparseimage'
        os.system('bunzip2 -k -c %s > %s' % (os.path.join(SNOWBERRY_DIR, 'template-image/template.sparseimage.bz2'),
                                             templateFile))
        remkdir('imaging')
        os.system('hdiutil attach %s -noautoopen -quiet -mountpoint imaging' % templateFile)
    
        remove('imaging/Doomsday.pkg') # included in bzipped image
        duptree('Doomsday Engine.app', 'imaging/Doomsday Engine.app')
        duptree('Doomsday Shell.app',  'imaging/Doomsday Shell.app')
        shutil.copy(os.path.join(DOOMSDAY_DIR, "doc/output/Read Me.rtf"), 'imaging/Read Me.rtf')

        volumeName = "Doomsday Engine " + DOOMSDAY_VERSION_FULL
        os.system('/usr/sbin/diskutil rename ' + os.path.abspath('imaging') + ' "' + volumeName + '"')
    
        os.system('hdiutil detach -quiet imaging')
        os.system('hdiutil convert %s -format UDZO -imagekey zlib-level=9 -o "../releases/%s"' % (
                templateFile, output_filename('_apps-' + mac_osx_suffix() + '.dmg')))
        remove(templateFile)

    def codesign(fn, opts=''):
        os.system('codesign --verbose -s "Developer ID Application: Jaakko Keranen" %s "%s"' % (opts, fn))
        
    def fw_codesign(app):
        print 'Signing libraries in %s...' % app
        for f in glob.glob(app + '/Contents/Frameworks/*.dylib'):
            codesign(f)

        print 'Signing Qt frameworks in %s...' % app
        for f in glob.glob(app + '/Contents/Frameworks/Qt*.framework'):
            name = f[f.find('/Qt'):-10]
            codesign(f + name)
    
        print 'Signing Qt plugins in %s...' % app
        for f in glob.glob(app + '/Contents/PlugIns/*/*.dylib'):
            codesign(f)
    
    fw_codesign('Doomsday Engine.app/Contents/Doomsday.app')
    fw_codesign('Doomsday Shell.app')

    print 'Signing Doomsday.app...'
    codesign('Doomsday Engine.app/Contents/Doomsday.app/Contents/Frameworks/SDL2.framework/SDL2')
    codesign("Doomsday Engine.app/Contents/Doomsday.app")

    print 'Signing Doomsday Engine.app...'
    os.system('ln -fs Versions/2.5/Python "Doomsday Engine.app/Contents/Frameworks/Python.framework/Python"')
    codesign("Doomsday Engine.app/Contents/Frameworks/Python.framework")
    codesign("Doomsday Engine.app/Contents/Frameworks/libwx_macud-2.8.0.dylib")
    codesign("Doomsday Engine.app/Contents/MacOS/python")
    codesign("Doomsday Engine.app", opts='--resource-rules ../macx/ResourceRules-OmitCompiledPython.plist')

    print 'Signing Doomsday Shell.app...'
    codesign("Doomsday Shell.app")
    
    # Package the apps and create an installer package.
    os.system('mkdir package')
    duptree('Doomsday Engine.app', 'package/Doomsday Engine.app')
    os.system('pkgbuild --identifier net.dengine.doomsday.frontend.pkg --version %s ' % DOOMSDAY_VERSION_FULL + \
        '--install-location /Applications --root package Frontend.pkg')
    os.system('rm -rf package')

    os.system('mkdir package')
    duptree('Doomsday Shell.app', 'package/Doomsday Shell.app')
    os.system('pkgbuild --identifier net.dengine.doomsday.shell.pkg --version %s ' % DOOMSDAY_VERSION_FULL + \
        '--install-location /Applications --root package Shell.pkg')
    os.system('rm -rf package')
    
    os.system("sed 's/${Version}/%s/' < ../macx/Distribution.xml.in > Distribution.xml" % DOOMSDAY_VERSION_FULL)
    os.system('mkdir res')
    shutil.copy(os.path.join(DOOMSDAY_DIR, "doc/output/Read Me.rtf"), 'res/Read Me.rtf')
    shutil.copy('../macx/background.png', 'res/background.png')
    os.system('productbuild --distribution Distribution.xml --package-path . --resources res --sign "Developer ID Installer: Jaakko Keranen" Doomsday.pkg')
        
    print 'Creating disk:', target
        
    # Compress a disk image containing the installer package.
    masterPkg = target
    volumeName = "Doomsday Engine " + DOOMSDAY_VERSION_FULL
    templateFile = os.path.join(SNOWBERRY_DIR, 'template-image/template.sparseimage')
    if not os.path.exists(templateFile):
        print 'Template .sparseimage not found, trying to extract from compressed archive...'
        os.system('bunzip2 -k "%s.bz2"' % templateFile)
    shutil.copy(templateFile, 'imaging.sparseimage')
    remkdir('imaging')
    os.system('hdiutil attach imaging.sparseimage -noautoopen -quiet -mountpoint imaging')
    try:
        shutil.copy('Doomsday.pkg', 'imaging/Doomsday.pkg')
    except Exception, ex:
        print 'No installer available:', ex
    shutil.copy(os.path.join(DOOMSDAY_DIR, "doc/output/Read Me.rtf"), 'imaging/Read Me.rtf')

    volumeName = "Doomsday Engine " + DOOMSDAY_VERSION_FULL
    os.system('/usr/sbin/diskutil rename ' + os.path.abspath('imaging') + ' "' + volumeName + '"')

    os.system('hdiutil detach -quiet imaging; rmdir imaging')
    os.system('hdiutil convert imaging.sparseimage -format UDZO -imagekey zlib-level=9 -o "' + target + '"')
    remove('imaging.sparseimage')


def win_release():
    """The Windows release procedure."""
    
    PROD_DIR = os.path.join(LAUNCH_DIR, 'products')
    if not os.path.exists(PROD_DIR):
        print "Creating the products directory."
        os.mkdir(PROD_DIR)

    PROD_DATA_DIR = os.path.join(PROD_DIR, 'data')
    if not os.path.exists(PROD_DATA_DIR):
        print "Creating the products/data directory."
        os.mkdir(PROD_DATA_DIR)

    PROD_DOC_DIR = os.path.join(PROD_DIR, 'doc')
    if not os.path.exists(PROD_DOC_DIR):
        print "Creating the products/doc directory."
        os.mkdir(PROD_DOC_DIR)

    # Generate the Inno Setup configuration file.
    script = file('win32\setup.iss.template', 'rt').read()
    file('win32\setup.iss', 'wt').write(script
        .replace('${YEAR}', time.strftime('%Y'))
        .replace('${BUILD}', DOOMSDAY_BUILD)
        .replace('${VERSION}', DOOMSDAY_VERSION_FULL)
        .replace('${VERSION_PLAIN}', DOOMSDAY_VERSION_FULL_PLAIN)
        .replace('${OUTPUT_FILENAME}', output_filename()))

    # Execute the win32 release script.
    os.chdir('win32')
    if os.system('dorel.bat ' + DOOMSDAY_BUILD_NUMBER):
        raise Exception("Failure in the Windows release script.")


def linux_release():
    """The Linux release procedure."""
    
    os.chdir(LAUNCH_DIR)

    def clean_products():
        # Remove previously built deb packages.
        os.system('rm -f ../doomsday*.deb ../doomsday*.changes ../doomsday*.tar.gz ../doomsday*.dsc')
        os.system('rm -f doomsday-fmod*.deb doomsday-fmod*.changes doomsday-fmod*.tar.gz doomsday-fmod*.dsc')
        #os.system('rm -f dsfmod/fmod-*.txt')

    clean_products()

    # Check that the changelog exists.
    if not os.path.exists('debian/changelog'):
        os.system('dch --check-dirname-level=0 --create --package doomsday -v %s-%s "Initial release."' % (DOOMSDAY_VERSION_FULL_PLAIN, DOOMSDAY_BUILD))

    if os.system('linux/gencontrol.sh && dpkg-buildpackage -b'):
        raise Exception("Failure to build from source.")

    # Build dsFMOD separately.
    os.chdir('dsfmod')
    logSuffix = "%s-%s.txt" % (sys.platform, platform.architecture()[0])
    if os.system('LD_LIBRARY_PATH=`pwd`/../builddir/libcore:`pwd`/../builddir/liblegacy dpkg-buildpackage -b > fmod-out-%s 2> fmod-err-%s' % (logSuffix, logSuffix)):
        raise Exception("Failure to build dsFMOD from source.")
    shutil.copy(glob.glob('../doomsday-fmod*.deb')[0], OUTPUT_DIR)
    shutil.copy(glob.glob('../doomsday-fmod*.changes')[0], OUTPUT_DIR)    
    os.chdir('..')

    # Place the result in the output directory.
    shutil.copy(glob.glob('../doomsday*.deb')[0], OUTPUT_DIR)
    shutil.copy(glob.glob('../doomsday*.changes')[0], OUTPUT_DIR)
    clean_products()


def main():
    prepare_work_dir()
    find_version()

    print "Checking OS...",

    try:
        if sys.platform == "darwin":
            print "Mac OS X (%s)" % mac_os_version()
            mac_release()
        elif sys.platform == "win32":
            print "Windows"
            win_release()
        elif sys.platform == "linux2":
            print "Linux"
            linux_release()
        else:
            print "Unknown!"
            print "I don't know how to make a release on this platform."
            exit_with_error()
    except Exception, x:
        print "Creating the release failed!"
        print x
        exit_with_error()

    os.chdir(LAUNCH_DIR)
    print "Done."


if __name__ == '__main__':
    main()



 # -*- coding: utf-8 -*-
import os
from distutils.core import setup

setup(name='pyassimp',
      version='0.1',
      license='ISC',
      description='Python bindings for the Open Asset Import Library (ASSIMP)',
      url='http://assimp.sourceforge.net/',
      packages=['pyassimp'],
      data_files=[('share/pyassimp', ['README.md']),
                  ('share/examples/pyassimp', ['scripts/' + f for f in os.listdir('scripts/')])]
      )



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file preferences.py Preferences in the Defaults/Snowberry Settings

import ui, events, language, paths
import sb.widget.button as wb
import sb.widget.text as wt
import sb.widget.list as wl
import sb.widget.tab as wa
import sb.confdb as st
import sb.util.dialog


# List widget of custom addon paths.
pathList = None

prefTabs = None


def init():
    # Register a listener for detecting the completion of Snowberry
    # startup.
    events.addNotifyListener(handleNotify, ['populating-area',
                                            'addon-paths-changed'])

    # Listen for the About button.
    events.addCommandListener(handleCommand, ['show-addon-paths'])
    
    # Commands for the popup menu.
    ui.addMenuCommand(ui.MENU_TOOLS, 'show-snowberry-settings', group=ui.MENU_GROUP_APP)


def handleNotify(event):
    """Handle notifications."""

    global pathList
    global prefTabs

    if (event.hasId('populating-area') and 
        event.getAreaId() == 'general-options'):

        # Create the addon paths list.
        area = event.getArea()
        area.setExpanding(True)
        area.setBorder(1)
        area.setWeight(1)
        
        # Tabs for different kinds of settings.
        tabs = area.createTabArea('pref-tab', style=wa.TabArea.STYLE_BASIC)        
        prefTabs = tabs
        box = tabs.addTab('addon-paths')
        box.setBorder(10)

        #box = area.createArea(boxedWithTitle='addon-paths', border=3)
        box.setWeight(1)
        #box.createText('addon-paths-info')
        box.setBorderDirs(ui.BORDER_NOT_BOTTOM)
        pathList = box.createList('addon-paths-list', wl.List.STYLE_MULTIPLE)
        box.setBorderDirs(ui.BORDER_NOT_TOP)
        pathList.setMinSize(100, 90)

        # Insert the current custom paths into the list.
        for p in paths.getAddonPaths():
            pathList.addItem(p)

        def addAddonPath():
            selection = sb.util.dialog.chooseFolder('addon-paths-add-prompt', '')
            if selection:
                pathList.deselectAllItems()
                pathList.addItem(selection)
                pathList.selectItem(selection)
                paths.addAddonPath(selection)                

        def removeAddonPath():
            for selection in pathList.getSelectedItems():
                pathList.removeItem(selection)
                paths.removeAddonPath(selection)
        
        box.setWeight(0)
        commands = box.createArea(alignment=ui.ALIGN_HORIZONTAL, border=2)
        commands.setWeight(0)

        # Button for adding new paths.
        button = commands.createButton('new-addon-path', wb.Button.STYLE_MINI)
        button.addReaction(addAddonPath)
        
        # Button for removing a path.
        button = commands.createButton('delete-addon-path',
                                       wb.Button.STYLE_MINI)
        button.addReaction(removeAddonPath)
        
        commands.setExpanding(False)
        commands.setWeight(1)
        commands.createText('refresh-required', align=wt.Text.RIGHT
                            ).setSmallStyle()
        
        # Checkboxes for hiding parts of the UI.
        #box = area.createArea(boxedWithTitle='ui-parts', border=3)
        box = tabs.addTab('ui-parts')
        box.setBorder(10)
        box.setWeight(0)

        # Create widgets for all (boolean) system settings.
        toggles = st.getSystemSettings(lambda s: s.getType() == 'toggle')
        toggles.sort(key=lambda s: s.getId())
        for toggle in toggles:
            box.createSetting(toggle)

        box.setWeight(1)
        box.addSpacer()
        box.setWeight(0)
        box.setBorder(10)
        box.createText('restart-required', align=wt.Text.RIGHT).setSmallStyle()

    elif event.hasId('addon-paths-changed'):
        # Insert the current custom paths into the list.
        pathList.clear()
        for p in paths.getAddonPaths():
            pathList.addItem(p)


def handleCommand(event):
    if event.hasId('show-addon-paths'):
        events.send(events.Command('show-snowberry-settings'))
        prefTabs.selectTab('addon-paths')
        pathList.focus()



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file main.py Addons Tab
##
## This plugin handles commands and notifications from the addons
## tree.  It also implements the load order dialog, the addon
## inspector, and the addon installation dialog.

import os, time, string
import paths, events, ui, language, widgets
import sb.util.dialog
import sb.widget.button as wg
import sb.widget.list as wl
import sb.widget.text as wt
import sb.profdb as pr
import sb.confdb as st
import sb.aodb as ao
import sb.addon
import tab30.installer
import tab30.loadorder
import tab30.inspector
import wx

ADDONS = 'tab-addons'

# True, if the Addons tab is currently visible.
tabVisible = False

# By default, sort by name.
listSortMode = 'name'

# True, if the addons list should be refreshed.
mustRefreshList = True


def init():
    # Manager for the addon list icons.
    global iconManager, addonIcons
    iconManager = widgets.IconManager(16, 16)
    addonIcons = [iconManager.get('unchecked'),
                  iconManager.get('checked'),
                  iconManager.get('defcheck')]

    # Create the Addons page.
    area = ui.createTab(ADDONS)

    WEIGHTS = [2, 3]
    
    # Top area for the filter controls.
    area.setWeight(0)
    topArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)
    topArea.setExpanding(False)

    # Counter text.
    global countText
    topArea.setWeight(WEIGHTS[0])
    countText = topArea.createText('', align=wt.Text.LEFT)   
    
    # Filter selection.
    topArea.setWeight(WEIGHTS[1])
    filterArea = topArea.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)
    filterArea.setWeight(0)
    filterArea.setExpanding(False)
    global listFilter
    filterArea.createText('addon-list-filter').resizeToBestSize()
    filterArea.setWeight(1)
    filterArea.setBorder(4, ui.BORDER_LEFT)
    listFilter = filterArea.createDropList('addon-list-filter-mode')
    listFilter.addItem('addon-list-filter-mode-compatible')
    listFilter.addItem('addon-list-filter-mode-compatible-pwad')
    listFilter.addItem('addon-list-filter-mode-pwad')
    listFilter.addItem('addon-list-filter-mode-all')
    listFilter.addItem('addon-list-filter-mode-all-with-boxes')
    listFilter.selectItem('addon-list-filter-mode-compatible')

    # Middle area for the category tree and the addon list.
    area.setWeight(1)
    area.setBorderDirs(ui.BORDER_LEFT_RIGHT)
    middleArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)
    area.setBorderDirs(ui.BORDER_NOT_TOP)

    # On the left, there is the main addon database controls.
    middleArea.setWeight(WEIGHTS[0])
    # Category tree.
    global tree
    middleArea.setBorder(8, ui.BORDER_RIGHT)
    tree = middleArea.createTree('category-tree')
    middleArea.setBorder(0)
    # Add all the categories into the tree.
    refreshCategories()
    
    # On the right, there is the filtered addon listing.
    middleArea.setWeight(WEIGHTS[1])
    global addonList 
    addonList = middleArea.createList('addon-list', style=wl.List.STYLE_COLUMNS)
    addonList.setImageList(iconManager.getImageList())
    addonList.setPopupMenuId('addon-list-popup')
    
    # Setup the columns.
    addonList.addColumn('addon-list-check', 20)
    addonList.addColumn('addon-list-name')
    addonList.addColumn('addon-list-version', 50)
    
    # Button arae in the bottom.
    area.setWeight(0)
    buttonArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=4)

    # Install (+) button for installing new addons.
    buttonArea.setWeight(0)
    buttonArea.setBorderDirs(ui.BORDER_TOP | ui.BORDER_RIGHT)
    buttonArea.createButton('install-addon', wg.Button.STYLE_MINI)

    global uninstallButton
    uninstallButton = buttonArea.createButton('uninstall-addon',
                                              wg.Button.STYLE_MINI)
    uninstallButton.disable()

    # The Refresh button reloads all addons.
    button = buttonArea.createButton('refresh-addon-database', wg.Button.STYLE_MINI)
    button.resizeToBestSize()   
    
    # Create addon listing controls.
    buttonArea.setWeight(1)
    buttonArea.addSpacer()
    buttonArea.setWeight(0)

    global infoButton
    infoButton = buttonArea.createButton('addon-info')
    infoButton.disable()

    global settingsButton
    settingsButton = buttonArea.createButton('addon-settings')
    settingsButton.disable()
    
    buttonArea.setBorderDirs(ui.BORDER_TOP)
    buttonArea.createButton('show-addon-paths')    
   
    ui.addMenuCommand(ui.MENU_TOOLS, 'load-order', group=ui.MENU_GROUP_AODB)

    # Registering a notification listener.
    events.addNotifyListener(handleNotification, ['tab-selected', 
                                                  'addon-list-icon-click',
                                                  'active-profile-changed',
                                                  'addon-list-popup-update-request',
                                                  'addon-attached',
                                                  'addon-detached',
                                                  'category-tree-selected',
                                                  'addon-installed',
                                                  'addon-database-reloaded',
                                                  'addon-list-selected',
                                                  'addon-list-deselected',
                                                  'addon-list-filter-mode-selected',
                                                  'addon-list-check-column-click',
                                                  'addon-list-name-column-click',
                                                  'addon-list-version-column-click'])

    # Registering a command listener to handle the Command events sent
    # by the buttons created above.
    events.addCommandListener(handleCommand, ['install-addon',
                                              'uninstall-addon',
                                              'addon-info',
                                              'addon-settings',
                                              'load-order',
                                              'addon-list-check-selected',
                                              'addon-list-uncheck-selected',
                                              'addon-list-check-all',
                                              'addon-list-uncheck-all',
                                              'addon-show-parent-box',
                                              'addon-show-box-category',
                                              'refresh-addon-database'])
                                              
    # Changing the current addon in the tree is done using commands with
    # the name of the addon as an identifier.
    events.addCommandListener(handleAnyCommand)
    
    # Menu commands.
    ui.addMenuCommand(ui.MENU_TOOLS, 'install-addon', pos=0, group=ui.MENU_GROUP_AODB)
    
         
def refreshCategories():
    """Recreates the categories of the category tree."""
    tree.freeze()
    oldSel = tree.getSelectedItem()
    expanded = tree.getExpandedItems()
    tree.clear()
    # Start populating from the root category.
    buildCategories(tree, 'category-tree-root', ao.getRootCategory())
    tree.expandItem('category-tree-root')
    map(tree.expandItem, expanded)        
    tree.selectItem(oldSel)
    tree.unfreeze()


def buildCategories(tree, parentId, parentCategory):
    """Builds categories recursively."""

    for cat in ao.getSubCategories(parentCategory):
        id = cat.getLongId()
        tree.addItem(id, parentId)
        buildCategories(tree, id, cat)    
        
    tree.sortItemChildren(parentId)
        

def refreshListIfVisible():
    global mustRefreshList
    mustRefreshList = True
    if tabVisible:
        refreshList()
        
        
def determineIcon(addon, profile, defaultAddons, profileAddons):
    """Determines which icon to use for an addon.
    
    @param addon  addon.Addon.
    @param profile  profile.Profile.
    @param defaultAddons  List of addon identifiers from Defaults.
    @param profileAddons  List of addon identifiers from the profile.
    
    @return Icon index.
    """
    ident = addon.getId()
    icon = addonIcons[0]
    if profile is pr.getDefaults():
        # In the defaults profile, addons are either checked or unchecked.
        checked = ident in defaultAddons

        # Invert the check mark on opt-out addons.
        if addon.isInversed():
            checked = not checked

        if checked:
            icon = addonIcons[1]
        else:
            icon = addonIcons[0]
    else:
        # In other profiles, addons may have the defcheck as well.
        checked = ident in profileAddons
        
        # Being in the Defaults inverts the selection.
        if ident in defaultAddons:
            checked = not checked
            inDefaults = True
        else:
            inDefaults = False
            
        if addon.isInversed():
            checked = not checked
            
        if checked:
            if inDefaults:
                icon = addonIcons[2]
            else:
                icon = addonIcons[1]
        else:
            icon = addonIcons[0]
    return icon
        

def refreshList():
    """Refreshes the list of addons."""
    addonList.freeze()

    # Try to keep the old selection, if there is one.
    oldSelections = addonList.getSelectedItems()

    addonList.clear()
    
    # Filtering parameters.
    profile = pr.getActive()
    isDefaults = profile is pr.getDefaults()
    if tree.getSelectedItem():
        filterCategory = ao.getCategory(tree.getSelectedItem())
    else:
        filterCategory = None
        
    # Which addons are currently attached?
    defaultAddons = pr.getDefaults().getAddons()
    if not isDefaults:
        profileAddons = pr.getActive().getAddons()
    else:
        profileAddons = None
    
    # Categories that will be in bold font in category-tree.
    boldCategories = []
    
    # Determine filtering options.
    mode = listFilter.getSelectedItem()
    onlyCompatible = False
    onlyPWAD = False
    alsoBoxes = False
    if 'compatible' in mode:
        onlyCompatible = True
    if 'pwad' in mode:
        onlyPWAD = True
    if 'with-boxes' in mode:
        alsoBoxes = True
        
    for addon in ao.getAddons():
        # Does this addon pass the filter?
        if not isDefaults and onlyCompatible:
            if not addon.isCompatibleWith(pr.getActive()):
                # Cannot be listed.
                continue
        if onlyPWAD and (addon.getType() != 'addon-type-wad' or not addon.isPWAD()):
            continue
        # Has a category been selected?
        if filterCategory and not filterCategory.isAncestorOf(addon.getCategory()):
            # Not in the category.
            continue
        # Addons in boxes are not listed unless the correct category is selected.
        if not alsoBoxes and addon.getBox():
            # Addons in boxes are only visible if the selected category is a box category.
            boxCateg = addon.getBox().getContentCategoryLongId()
            if not filterCategory: continue
            if filterCategory.getLongId().find(boxCateg) != 0:
                # Not the right category.
                continue
        # Passed the filter!
        id = addon.getId()
        versionStr = ''
        icon = determineIcon(addon, profile, defaultAddons, profileAddons)
            
        if language.isDefined(id + '-version'):
            versionStr = language.translate(id + '-version')
            
        name = language.translate(id)
        if addon.getBox() and alsoBoxes:
            name += ' ' + language.translate('addon-list-in-box')
            
        addonList.addItemWithColumns(id,  icon, name, versionStr)
      
    # Reselect old items.
    oldSelections.reverse()
    for sel in oldSelections:
        addonList.selectItem(sel)
        addonList.ensureVisible(sel)

    sortList()
    updateButtons()

    addonList.unfreeze()
    
    # Update the counter text.
    countText.setText(language.expand(language.translate('addon-counter'),
                                      str(len(addonList.getItems())), 
                                      str(ao.getAddonCount())))
    
    # Update boldings in category-tree.
    # --TODO!--   
    
    global mustRefreshList
    mustRefreshList = False
    
    
def refreshItemInList(item):
    addon = ao.get(item)
    defaultAddons = pr.getDefaults().getAddons()
    if pr.getActive() is not pr.getDefaults():
        profileAddons = pr.getActive().getAddons()
    else:
        profileAddons = None
    
    # Update the image of the item.
    addonList.setItemImage(item, determineIcon(addon, pr.getActive(),
                                               defaultAddons, profileAddons))
                                               
                                               
def sortList():
    """Sort the current addon list according to the sort mode."""
    
    # Unsorted items of the list. Sorting will be done based on this information.
    if listSortMode == 'check':
        icons = addonList.getItemImages()
    items = addonList.getItems()
    keys = map(lambda i: (language.translate(i).lower(), 
                          language.translate(i + '-version', '0.0')), items)
    
    def checkSortCmp(a, b):
        # First compare check icon (descending).
        result = cmp(icons[b], icons[a])
        if result != 0:
            return result
        # Then compare as in by name.
        return nameSortCmp(a, b)

    def nameSortCmp(a, b):
        # First by name.
        result = cmp(keys[a][0], keys[b][0])
        if result != 0:
            return result
        # Then by version (descending).
        return cmp(keys[b][1], keys[a][1])

    def versionSortCmp(a, b):
        # First by version (descending).
        result = cmp(keys[b][1], keys[a][1])
        if result != 0:
            return result
        # Then by name.
        return cmp(keys[a][0], keys[b][0])
        
    funcs = {'version': versionSortCmp, 'check': checkSortCmp, 'name': nameSortCmp}
    addonList.sortItems(funcs[listSortMode])
                              

def handleNotification(event):
    """This is called when someone sends a notification event.

    @param event An events.Notify object.
    """
    global tabVisible
    global mustRefreshList

    if event.hasId('addon-list-icon-click'):
        # Clicking on the list icon will attach/detach the addon.
        profile = pr.getActive()
        if event.item in profile.getUsedAddons():
            profile.dontUseAddon(event.item)
        else:
            profile.useAddon(event.item)
            
    elif event.hasId('addon-attached') or event.hasId('addon-detached'):
        refreshItemInList(event.getAddon())

    elif event.hasId('addon-list-filter-mode-selected'):
        refreshListIfVisible()

    elif event.hasId('addon-list-check-column-click') or \
         event.hasId('addon-list-name-column-click') or \
         event.hasId('addon-list-version-column-click'):
        global listSortMode
        if 'check' in event.getId():
            listSortMode = 'check'
        elif 'version' in event.getId():
            listSortMode = 'version'
        else:
            listSortMode = 'name'
        sortList()

    elif event.hasId('tab-selected'):
        # If there is a refresh pending, do it now that the tab has 
        # been selected.
        if event.getSelection() == ADDONS:
            tabVisible = True
            if mustRefreshList:
                refreshList()
        else:
            tabVisible = False
    
    elif event.hasId('active-profile-changed'):
        refreshListIfVisible()
            
        # Fill the tree with an updated listing of addons.
        #tree.populateWithAddons(pr.getActive())
        #settingsButton.disable()

    elif event.hasId('addon-list-popup-update-request'):
        # The addon list is requesting an updated popup menu.
        # Menu items depend on which items are selected.
        items = addonList.getSelectedItems()
        menu = []
        
        if len(items) == 1:
            # One addon is selected.
            if ao.get(items[0]).getBox():
                menu += ['addon-show-parent-box']
            if ao.get(items[0]).getType() == 'addon-type-box':
                menu += ['addon-show-box-category']
            # Info is always available for all addons.
            menu += ['addon-info'] 
            if st.haveSettingsForAddon(items[0]):
                menu += ['addon-settings']
        elif len(items) > 1:
            # Multiple addons selected.
            menu += ['addon-list-check-selected',
                     'addon-list-uncheck-selected']

        if len(items) > 0: menu += ['uninstall-addon']

        # Append the common items.
        if len(menu) > 0: menu.append('-')
        menu += ['install-addon'] 
        menu += ['load-order']

        menu += ['-', 'addon-list-check-all', 'addon-list-uncheck-all']
        
        addonList.setPopupMenu(menu)

    elif event.hasId('category-tree-selected'):
        refreshListIfVisible()
        # Enable the settings button if there are settings for this
        # addon.
        #if st.haveSettingsForAddon(event.getSelection()):
        #    settingsButton.enable()
        #else:
        #    settingsButton.disable()

        # Can this be uninstalled?
        #uninstallButton.enable(ao.exists(event.getSelection()) and
        #                       ao.get(event.getSelection()).getBox() == None)

    elif event.hasId('addon-list-selected') or event.hasId('addon-list-deselected'):
        updateButtons()

    elif event.hasId('addon-installed'):
        refreshCategories()
        refreshListIfVisible()
        #tree.selectAddon(event.addon)

    elif event.hasId('addon-database-reloaded'):
        refreshCategories()
        refreshListIfVisible()
        
        
def updateButtons():
    items = addonList.getSelectedItems()
    if len(items) == 0:
        settingsButton.disable()
        uninstallButton.disable()
        infoButton.disable()
        return
        
    if len(items) > 1:
        # More than one.
        settingsButton.disable() 

        canUninstall = False
        # At least one uninstallable?
        for addon in items:
            if ao.isUninstallable(addon):
                uninstallButton.enable()
                canUninstall = True
                break
        if not canUninstall:
            uninstallButton.disable()

        infoButton.disable()
        return

    sel = items[0]
    settingsButton.enable(st.haveSettingsForAddon(sel))
    infoButton.enable()
        
    # Can this be uninstalled?
    uninstallButton.enable(ao.isUninstallable(sel))


def showInAddonList(addon):
    if ao.exists(addon):
        if addon in addonList.getItems():
            # Select the addon in the tree.
            addonList.deselectAllItems()
            addonList.selectItem(addon)
            addonList.ensureVisible(addon)
        else:
            # The addon was not currently visible, so make sure it is.
            tree.selectItem('category-tree-root')
            addonList.deselectAllItems()
            refreshListIfVisible()
            addonList.selectItem(addon)
            addonList.ensureVisible(addon)


def handleAnyCommand(event):
    """Handle any command. This is called whenever a command is broadcasted.
    Must not do anything slow."""
    # Is the command an addon identifier?
    showInAddonList(event.getId())
        

def handleCommand(event):
    """This is called when someone sends a command event.

    @param event An events.Command object."""

    # Figure out which addon has been currently selected in the addons
    # tree.  The action will target this addon.
    selected = ''

    if event.hasId('refresh-addon-database'):
        ao.refresh()

    elif event.hasId('install-addon'):
        tab30.installer.run()

    elif event.hasId('uninstall-addon'):
        selectedItems = addonList.getSelectedItems()
                
        # Only take the uninstallable addons.
        addons = filter(lambda i: ao.isUninstallable(i), selectedItems)
        if len(addons) == 0:
            return
            
        # Make sure the user wants to uninstall.
        dialog, area = sb.util.dialog.createButtonDialog(
            'uninstall-addon-dialog',
            ['no', 'yes'], 'no', resizable=False)

        if len(addons) == 1:
            text = language.translate('uninstall-addon-query')
            area.setExpanding(True)
            area.createRichText(language.expand(text, language.translate(addons[0])))
        else:
            area.setWeight(0)
            area.createText('uninstall-addon-query-several')
            msg = '<html><ul>'
            for addon in addons:
                msg += '<li>' + language.translate(addon) + '</li>'
            msg += '</ul></html>'
            ft = area.createFormattedText()
            ft.setText(msg)
            ft.setMinSize(400, 150)
    
        if dialog.run() == 'yes':
            for addon in addons:
                ao.uninstall(addon)
            ao.refresh()

    elif event.hasId('addon-info'):
        sel = addonList.getSelectedItems()
        if len(sel) > 0:
            # Show the Addon Inspector.
            tab30.inspector.run(ao.get(sel[0]))

    elif event.hasId('addon-settings'):
        sel = addonList.getSelectedItems()
        if len(sel) > 0:
            command = events.Command('show-addon-settings')
            command.addon = sel[0]
            events.send(command)        

    elif event.hasId('load-order'):
        # Open the Load Order dialog.
        tab30.loadorder.run(pr.getActive())

    elif event.hasId('addon-list-check-selected'):
        for addon in addonList.getSelectedItems():
            pr.getActive().useAddon(addon)

    elif event.hasId('addon-list-uncheck-selected'):
        for addon in addonList.getSelectedItems():
            pr.getActive().dontUseAddon(addon)

    elif event.hasId('addon-list-check-all'):
        for addon in addonList.getItems():
            pr.getActive().useAddon(addon)

    elif event.hasId('addon-list-uncheck-all'):
        for addon in addonList.getItems():
            pr.getActive().dontUseAddon(addon)
            
    elif event.hasId('addon-show-parent-box'):
        for addon in addonList.getSelectedItems():
            a = ao.get(addon)
            if a.getBox():
                showInAddonList(a.getBox().getId())
            break
            
    elif event.hasId('addon-show-box-category'):
        for addon in addonList.getSelectedItems():
            a = ao.get(addon)
            if a.getType() == 'addon-type-box':
                addonList.deselectAllItems()
                tree.selectItem(a.getContentCategoryLongId())
                refreshListIfVisible()
            break

    #elif event.hasId('expand-all-categories'):
    #    tree.expandAllCategories(True)

    #elif event.hasId('collapse-all-categories'):
    #    tree.expandAllCategories(False)

    #elif event.hasId('check-category'):
    #    tree.checkSelectedCategory(True)

    #elif event.hasId('uncheck-category'):
    #    tree.checkSelectedCategory(False)



def oldInit():
    # Create the Addons page.
    area = ui.createTab(ADDONS)
    area.setBorderDirs(ui.BORDER_NOT_BOTTOM)
    
    # Let's create the addons tree...
    area.setWeight(1)

    global tree
    tree = area.createTree()

    # Add all the categories into the tree.
    tree.createCategories()

    # The addon control buttons.
    area.setWeight(0)
    area.setBorderDirs(ui.BORDER_NOT_TOP)
    buttonArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=4)

    buttonArea.setWeight(0)

    # Install (+) button for installing new addons.
    buttonArea.setBorderDirs(ui.BORDER_TOP | ui.BORDER_RIGHT)
    buttonArea.createButton('install-addon', wg.Button.STYLE_MINI)

    global uninstallButton
    uninstallButton = buttonArea.createButton('uninstall-addon',
                                              wg.Button.STYLE_MINI)
    uninstallButton.disable()

    # The Refresh button reloads all addons.
    button = buttonArea.createButton('refresh-addon-database', wg.Button.STYLE_MINI)
    button.resizeToBestSize()   

    buttonArea.setWeight(3)
    buttonArea.addSpacer()
    buttonArea.setWeight(0)

    global settingsButton
    settingsButton = buttonArea.createButton('addon-settings')
    settingsButton.disable()
    
    buttonArea.setBorderDirs(ui.BORDER_TOP)
    buttonArea.createButton('show-addon-paths')

    ui.addMenuCommand(ui.MENU_TOOLS, 'load-order')

    # Registering a notification listener.
    events.addNotifyListener(handleNotification, ['active-profile-changed',
                                                  'addon-popup-request',
                                                  'addon-installed',
                                                  'addon-database-reloaded'])

    # Registering a command listener to handle the Command events sent
    # by the buttons created above.
    events.addCommandListener(handleCommand, ['install-addon',
                                              'uninstall-addon',
                                              'addon-info',
                                              'addon-settings',
                                              'load-order',
                                              'expand-all-categories',
                                              'collapse-all-categories',
                                              'check-category',
                                              'uncheck-category',
                                              'refresh-addon-database'])
                                              
    # Changing the current addon in the tree is done using commands with
    # the name of the addon as an identifier.
    events.addCommandListener(handleAnyCommand)         



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file ui.py User Interface
##
## This module contains classes and functions for managing the
## application's user interface.  The basic concept is that the
## interface has been divided into a number of areas.  The top-level
## areas (title, profiles, command, help) are created when the main
## window is constructed, and cannot be changed at runtime.  The
## contents of each area can be modified at runtime.
##
## Many of the areas are managed by plugins.  This means the plugin
## creates all the widgets inside the area.
##
## Each area can be divided into further sub-areas, which are treated
## the same way as the top-level areas.  The Area class can represent
## either a top-level area or a sub-area.

import sys, os, wx, string
import host, events, widgets, language, paths
import sb.widget.tab
import sb.profdb as pr
import sb.confdb as st
import logger
from widgets import uniConv

# Default areas.
TITLE = 'main-title'
PROFILES = 'main-profiles'
COMMAND = 'main-command'
PREFCOMMAND = 'main-pref-command'
TABS = 'main-tabs'
HELP = 'main-help'

# Menus.
MENU_APP = 0
MENU_PROFILE = 1
MENU_TOOLS = 2
MENU_HELP = 3

# Menu groups.
MENU_GROUP_LAUNCH = 10
MENU_GROUP_ADDON = 20
MENU_GROUP_AODB = 30
MENU_GROUP_PROFILE = 40
MENU_GROUP_PROFDB = 50
MENU_GROUP_APP = 60

# Layout alignments.
ALIGN_HORIZONTAL = 0
ALIGN_VERTICAL = 1

# Border flags.
BORDER_TOP = 1
BORDER_RIGHT = 2
BORDER_BOTTOM = 4
BORDER_LEFT = 8
BORDER_TOP_BOTTOM = (1 | 4)
BORDER_LEFT_RIGHT = (2 | 8)
BORDER_NOT_TOP = (2 | 4 | 8)
BORDER_NOT_RIGHT = (1 | 4 | 8)
BORDER_NOT_BOTTOM = (1 | 2 | 8)
BORDER_NOT_LEFT = (1 | 2 | 4)
BORDER_ALL = (1 | 2 | 4 | 8)

# Optional areas.
USE_TITLE_AREA = not st.getSystemBoolean('main-hide-title')
USE_HELP_AREA = not st.getSystemBoolean('main-hide-help')
USE_MINIMAL_PROFILE = st.getSystemBoolean('profile-minimal-mode')

# A Windows kludge: background colour for tabs and the controls in them.
#tabBgColour = wx.Colour(250, 250, 250)
#st.getSystemInteger('tab-background-red'),
#                        st.getSystemInteger('tab-background-green'),
#                        st.getSystemInteger('tab-background-blue'))

# An array of UI areas.
uiAreas = {}
mainPanel = None

# Items for the popup menu. Up to seven menus.
menuItems = [[], [], [], [], [], [], []]


class Timer (wx.Timer):
    """A timer class."""

    def __init__(self, notifyId=''):
        """Initialize the timer."""
        wx.Timer.__init__(self)
        self.notifyId = notifyId

    def Notify(self):
        """Called when the timer expires."""
        self.expire()

    def start(self, milliSeconds):
        """Start the timer in one-shot mode.

        @param milliSeconds  Milliseconds after which the timer expires.
        """
        self.Start(milliSeconds, wx.TIMER_ONE_SHOT)

    def expire(self):
        """Send a notification if one has been defined. This method can be
        overridden by subclasses.
        """
        if self.notifyId:
            events.send(events.Notify(self.notifyId))


class MainPanel (wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, -1, style=wx.NO_BORDER |
                          wx.CLIP_CHILDREN)

        from sb.widget.area import Area

        aSizer = wx.BoxSizer(wx.HORIZONTAL)
        bSizer = wx.BoxSizer(wx.VERTICAL)
        cSizer = wx.BoxSizer(wx.HORIZONTAL)
        dSizer = wx.BoxSizer(wx.VERTICAL)
        eSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.verticalSizer = bSizer

        # Title area.
        if USE_TITLE_AREA:
            titlePanel = wx.Panel(self, -1,
                                  style=wx.NO_BORDER | wx.CLIP_CHILDREN)
            _newArea( Area(TITLE, titlePanel, ALIGN_HORIZONTAL) )
            bSizer.Add(titlePanel, 0, wx.EXPAND)

        # Profile area.
        tabPanel = wx.Panel(self, -1,
                            style=wx.NO_BORDER | wx.CLIP_CHILDREN)
        tabsArea = Area(TABS, tabPanel, ALIGN_VERTICAL, 16)
        _newArea(tabsArea)

        dSizer.Add(tabPanel, 5, wx.EXPAND | wx.ALL, 3)

        # Command area.
        commandPanel = wx.Panel(self, -1, style=wx.NO_BORDER
                                 | wx.CLIP_CHILDREN)
        area = Area(COMMAND, commandPanel, ALIGN_HORIZONTAL, 18,
                    borderDirs=BORDER_LEFT | BORDER_RIGHT | BORDER_BOTTOM)
        # The command area is aligned to the right.
        area.addSpacer()
        area.setExpanding(False)
        area.setWeight(0)
        _newArea(area)

        # Preferences command area.
        prefCommandPanel = wx.Panel(self, -1, style=wx.NO_BORDER |
                                    wx.CLIP_CHILDREN)
        area = Area(PREFCOMMAND, prefCommandPanel,
                    ALIGN_HORIZONTAL, 18,
                    borderDirs=BORDER_LEFT | BORDER_RIGHT | BORDER_BOTTOM)
        area.setExpanding(False)
        area.setWeight(0)
        _newArea(area)

        eSizer.Add(prefCommandPanel, 0, wx.EXPAND)
        eSizer.Add(commandPanel, 1, wx.EXPAND)
        dSizer.Add(eSizer, 0, wx.EXPAND)
        cSizer.Add(dSizer, 5, wx.EXPAND) # notebook
        bSizer.Add(cSizer, 1, wx.EXPAND)
        aSizer.Add(bSizer, 7, wx.EXPAND)

        # Help area.
        self.SetSizer(aSizer)
        self.SetAutoLayout(True)

    def getTabs(self):
        """Returns the TabArea widget that contains the main tabbing
        area."""
        return self.tabs

    def updateLayout(self):
        self.verticalSizer.Layout()


INITIAL_SASH_POS = 180
INITIAL_PROFILE_SASH_POS = 190

if st.isDefined('main-split-position'):
    INITIAL_SASH_POS = st.getSystemInteger('main-split-position')

if st.isDefined('main-profile-split-position'):
    INITIAL_PROFILE_SASH_POS = st.getSystemInteger(
        'main-profile-split-position')


class MainFrame (wx.Frame):
    """The main frame is the main window of Snowberry."""

    def __init__(self, title):
        """Initialize the main window.

        @param title  Title for the main window.
        """
        from sb.widget.area import Area

        # Commands for the popup menu.
        self.menuCommandMap = {}

        # FIXME: Layout is botched. Way too wide.
        if host.isMac():
            initialSize = (769, 559)
        elif host.isUnix():
            initialSize = (900, 610)
        else:
            initialSize = (733, 527)

        # The configuration may define a window size.
        if st.isDefined('main-width'):
            initialSize = (st.getSystemInteger('main-width'), initialSize[1])
        if st.isDefined('main-height'):
            initialSize = (initialSize[0], st.getSystemInteger('main-height'))

        # The configuration can also specify the window position.
        if st.isDefined('main-x') and st.isDefined('main-y'):
            initialPos = (st.getSystemInteger('main-x'), st.getSystemInteger('main-y'))
        else:
            initialPos = None

        wx.Frame.__init__(self, None, -1, title, size=initialSize)
        #self.SetExtraStyle(wx.FRAME_EX_METAL)
        #self.Create(None, -1, title, size=initialSize)

        if initialPos is not None:
            self.MoveXY(*initialPos)
        else:
            self.Center()

        # Set the icon for the frame.
        icon = wx.Icon('graphics/snowberry.ico', wx.BITMAP_TYPE_ICO)
        self.SetIcon(icon)

        #self.Iconize(True)
        #self.Hide()

        SPLITTER_ID = 9501
        PROF_SPLITTER_ID = 9502
        self.splitter = None
        self.profSplitter = None

        # The parentWin is where the profSplitter and the help panel
        # are inside.
        parentWin = self

        if USE_HELP_AREA:
            # The help area is in a splitter.
            self.splitter = wx.SplitterWindow(self, SPLITTER_ID,
                                              style=wx.SP_3DSASH)
            self.splitter.SetMinimumPaneSize(10)
            parentWin = self.splitter

        if not USE_MINIMAL_PROFILE:
            self.profSplitter = wx.SplitterWindow(parentWin, PROF_SPLITTER_ID,
                                                  style=wx.SP_3DSASH)# |
                                                  #wx.SP_NO_XP_THEME)
            self.profSplitter.SetMinimumPaneSize(100)

            # Profile panel.
            profilePanel = wx.Panel(self.profSplitter, -1, style=wx.NO_BORDER
                                    | wx.CLIP_CHILDREN)
            area = Area(PROFILES, profilePanel, ALIGN_VERTICAL, 10)
            _newArea(area)

            # Create panels inside the profile splitter.
            self.mainPanel = MainPanel(self.profSplitter)
        else:
            profilePanel = None

            self.mainPanel = MainPanel(parentWin)

            getArea(TABS).setWeight(0)
            profArea = getArea(TABS).createArea(alignment=ALIGN_HORIZONTAL, border=12)
            profArea.setId(PROFILES)
            _newArea(profArea)
            getArea(TABS).setWeight(1)
            getArea(TABS).setBorderDirs(BORDER_NOT_TOP)

        # Create a TabArea into the TABS area.
        self.mainPanel.tabs = getArea(TABS).createTabArea(
            'tab', style=sb.widget.tab.TabArea.STYLE_BASIC)

        if st.isDefined('main-split-position'):
            self.splitPos = INITIAL_SASH_POS
        else:
            self.splitPos = None
        self.profSplitPos = INITIAL_PROFILE_SASH_POS

        # Create the help area.
        if self.splitter:
            self.helpPanel = wx.Panel(self.splitter, -1, style = wx.NO_BORDER
                                      | wx.CLIP_CHILDREN)
            self.helpPanel.SetBackgroundColour(wx.WHITE)
            _newArea( Area(HELP, self.helpPanel, ALIGN_VERTICAL,
                           border=4) )
            # Init the splitter.
            leftSide = self.profSplitter
            if not leftSide:
                leftSide = self.mainPanel
            self.splitter.SplitVertically(leftSide, self.helpPanel,
                                          -INITIAL_SASH_POS)
        else:
            self.helpPanel = None

        if self.profSplitter:
            self.profSplitter.SplitVertically(profilePanel, self.mainPanel,
                                              self.profSplitPos)

        # Listen for changes in the sash position.
        wx.EVT_SPLITTER_SASH_POS_CHANGED(self, SPLITTER_ID,
                                         self.onSplitChange)
        wx.EVT_SPLITTER_SASH_POS_CHANGED(self, PROF_SPLITTER_ID,
                                         self.onProfSplitChange)

        # The main panel should be globally accessible.
        global mainPanel
        mainPanel = self.mainPanel

        # Intercept the window close event.
        wx.EVT_CLOSE(self, self.onWindowClose)

        # Maintain the splitter position.
        wx.EVT_SIZE(self, self.onWindowSize)

        # Listen to some commands.
        events.addCommandListener(self.handleCommand, ['quit'])

        # Create a menu bar.
        self.menuBar = wx.MenuBar()

    def show(self):
        #if self.splitter:
        #    self.splitter.Thaw()
        #    self.splitter.Refresh()
        pass

    def updateLayout(self):
        """Update the layout of the widgets inside the main window."""

        # Also update all UI areas.
        for key in uiAreas:
            uiAreas[key].updateLayout()

        self.mainPanel.updateLayout()
        #self.Show()

        if not host.isMac():
            self.mainPanel.GetSizer().Fit(self.mainPanel)

        # The main panel's sizer does not account for the help panel.
        #if host.isUnix(): # or host.isWindows():
        #    windowSize = self.GetSizeTuple()
        #    self.SetSize((windowSize[0] + INITIAL_SASH_POS, windowSize[1]))

    def onWindowSize(self, ev):
        """Handle the wxWidgets event that is sent when the main window
        size changes.

        @param ev The wxWidgets event.
        """
        # Allow others to handle this event as well.
        ev.Skip()

        if self.splitter:
            if self.splitPos:
                pos = self.splitPos
            else:
                pos = INITIAL_SASH_POS

            self.splitter.SetSashPosition(ev.GetSize()[0] - pos)

    def onSplitChange(self, ev):
        """Update the splitter anchor position."""

        if self.splitter:
            self.splitPos = self.GetClientSizeTuple()[0] - ev.GetSashPosition()

    def onProfSplitChange(self, ev):
        """Update the profile splitter anchor position."""

        self.profSplitPos = ev.GetSashPosition()

    def onWindowClose(self, ev):
        """Handle the window close event that is sent when the user
        tries to close the main window.  Broadcasts a 'quit'
        events.Notify object so that all the listeners will know that
        the application is closing down.

        @param ev A wxCloseEvent object.
        """
        # Send a 'quit' command so that everyone has a chance to save
        # their current status.
        events.send(events.Notify('quit'))

        # Save window size to UserHome/conf/window.conf.
        winSize = self.GetSizeTuple()
        fileName = os.path.join(paths.getUserPath(paths.CONF), 'window.conf')
        try:
            f = file(fileName, 'w')
            f.write('# This file is generated automatically.\n')
            f.write('appearance main (\n')
            f.write('  width = %i\n  height = %i\n' % winSize)
            f.write('  x = %i\n  y = %i\n' % self.GetPositionTuple())
            if self.splitPos != None:
                f.write('  split-position = %i\n' % self.splitPos)
            f.write('  profile-split-position = %i\n' % self.profSplitPos)
            f.write(')\n')
        except:
            # Window size not saved.
            pass

        # Destroy the main frame.
        self.Destroy()
        return True

    def handleCommand(self, event):
        """This is called whenever someone broadcasts an
        events.Command object.  The main window is only interested in
        the 'quit' command, on the reception of which it closes
        itself.

        @param event An events.Command object.
        """
        if event.hasId('quit'):
            self.Close()

    def updateMenus(self):
        """Create the main frame menus based on menuItems array."""

        global menuItems
        self.menuCommandMap = {}

        # Array of menus. Number to wxMenu.
        self.menus = {}

        # Three levels of priority.
        for level in range(len(menuItems)):
            if len(menuItems[level]) == 0:
                # No menu for this.
                self.menus[level] = None
                continue

            self.menus[level] = wx.Menu()
            self.menuBar.Append(self.menus[level],
                                language.translate("menu-" + str(level)))

            # Sort the items based on groups, and append separators where necessary.
            menuItems[level].sort(lambda x, y: cmp(x[3], y[3]))
            separated = []
            previousItem = None
            for item in menuItems[level]:
                if previousItem and previousItem[3] != item[3]:
                    separated.append(('-', '', False, item[3]))
                separated.append(item)
                previousItem = item

            # Create the menu items.
            for itemId, itemCommand, itemSeparate, itemGroup in separated:
                if itemId == '-':
                    # This is just a separator.
                    self.menus[level].AppendSeparator()
                    continue

                if itemSeparate and self.menus[level].GetMenuItemCount() > 0:
                    self.menus[level].AppendSeparator()

                menuItemId = 'menu-' + itemId

                accel = ''
                if language.isDefined(menuItemId + '-accel'):
                    accel = "\t" + language.translate(menuItemId + '-accel')

                # Generate a new ID for the item.
                wxId = wx.NewId()
                self.menuCommandMap[wxId] = itemCommand
                self.menus[level].Append(wxId,
                                        uniConv(language.translate(menuItemId) + accel))
                wx.EVT_MENU(self, wxId, self.onPopupCommand)

                if host.isMac():
                    # Special menu items on Mac.
                    if itemId == 'about':
                        wx.App_SetMacAboutMenuItemId(wxId)
                    if itemId == 'quit':
                        wx.App_SetMacExitMenuItemId(wxId)
                    if itemId == 'show-snowberry-settings':
                        wx.App_SetMacPreferencesMenuItemId(wxId)

        if host.isMac():
            # Special Help menu on Mac.
            wx.App_SetMacHelpMenuTitleName(language.translate('menu-' + str(MENU_HELP)))

        self.SetMenuBar(self.menuBar)

    def onPopupCommand(self, ev):
        """Called when a selection is made in the popup menu."""
        events.send(events.Command(self.menuCommandMap[ev.GetId()]))

    def getMenuItem(self, identifier):
        """Finds a menu item.
        @param identifier  Identifier of the menu item.
        @return wxMenuItem, or None."""

        for item in self.menuCommandMap.keys():
            if self.menuCommandMap[item] == identifier:
                return self.menuBar.FindItemById(item)
        return None


class SnowberryApp (wx.App):
    def __init__(self):
        wx.App.__init__(self, 0)

    def OnInit(self):
        # Create and show the main window.
        self.mainFrame = MainFrame('Snowberry')

        self.SetTopWindow(self.mainFrame)
        return True

    def updateTitle(self):
        windowTitle = st.getSystemString('snowberry-title') + ' ' + \
                      st.getSystemString('snowberry-version')
        self.mainFrame.SetTitle(windowTitle)

    def showMainWindow(self):
        self.updateTitle()
        self.mainFrame.Thaw()
        self.mainFrame.Refresh()


def getMainPanel():
    """Returns the wxPanel of the main window."""
    return mainPanel


def addMenuCommand(level, identifier, label=None, pos=None, separate=False, group=''):
    """Insert a new popup menu item for the Menu button.

    @param level       Priority level (0, 1, 2).
    @param identifier  Identifier of the command.
    @param label       Label for the command. If not specified, the same as the identifier.
    @param pos         Position in the menu. None will cause the new item to be appended
                       to the end of the menu. The first item in the menu has position 0.
    @param separate    If True and there are existing items already, add a separator.
    @param group       If specified, items are sorted by this before adding to the menu.
                       The items without a group are inserted first. Groups are separated
                       with separators.
    """
    global menuItems

    if host.isMac() and level == 0:
        # On the Mac, the Snowberry menu is integrated into the app menu.
        # There is no need for a level zero menu.
        # Put the items temporarily on level 1, where they will be moved
        # by wxPython to the app menu.
        level = 1

    if label is None:
        item = (identifier, identifier, separate, group)
    else:
        item = (identifier, label, separate, group)

    if pos is None:
        menuItems[level].append(item)
    else:
        menuItems[level].insert(pos, item)


def isMenuEmpty(level):
    """Determines whether a certain menu is currently empty.
    @return True or False."""
    return len(menuItems[level]) == 0


def enableMenuCommand(identifier, enable=True):
    """Enables or disables a command in the menu.

    @param identifier  Identifier of the menu item.
    @param enable  True, if the item should be enabled. False, if disabled.
    """
    item = app.mainFrame.getMenuItem(identifier)
    if item:
        if enable:
            item.Enable(True)
        else:
            item.Enable(False)


def disableMenuCommand(identifier):
    enableMenuCommand(identifier, False)


def getArea(id):
    """Returns the UI area with the specified ID.

    @return An Area object.
    """
    return uiAreas[id]


def _newArea(area):
    """A private helper function for adding a new area to the list of
    identifiable areas.  Sub-areas are not identifiable.

    @param area An Area object.
    """
    global uiAreas
    uiAreas[area.getId()] = area


def selectTab(id):
    """Select a tab in the tab switcher.

    @param id Identifier of the tab.
    """
    mainPanel.getTabs().selectTab(id)


def createTab(id):
    """Create a new tab in the tab switcher.

    @param id The identifier of the new tab.  The label of the tab
    will be the translation of this identifier.

    @return An Area object that represents the new tab.
    """
    area = mainPanel.getTabs().addTab(id)
    area.setBorder(10)

    # Register this area so it can be found with getArea().
    _newArea(area)

    return area


def prepareWindows():
    """Layout windows and make everything ready for action."""

    events.send(events.Notify('preparing-windows'))

    app.mainFrame.updateMenus()
    app.mainFrame.updateLayout()
    app.mainFrame.Show()
    app.mainFrame.Freeze()


def startMainLoop():
    """Start the wxWidgets main loop."""

    # Before we give wxPython all control, update the layout of UI
    # areas.
    app.showMainWindow()

    # Display any issues that came up during init.
    logger.show()

    # Notify everyone that the main loop is about to kick in.
    events.send(events.Notify('init-done'))

    app.MainLoop()


def freeze():
    app.mainFrame.Freeze()


def unfreeze():
    app.mainFrame.Thaw()
    app.mainFrame.Refresh()


# Initialize the UI module by creating the main window of the application.
app = SnowberryApp()

# Construct an icon manager that'll take care of the icons used in the
# user interface.
widgets.initIconManager()



import os
import builder.config
import subprocess


def run_git(cmdLine, ignoreResult=False):
    result = subprocess.call(cmdLine, shell=True)
    if result and not ignoreResult:
        raise Exception("Failed to run git: " + cmdLine)


def git_checkout(ident):
    """Checkout the branch or tag @a ident from the repository."""
    print 'Checking out %s...' % ident
    os.chdir(builder.config.DISTRIB_DIR)
    run_git("git checkout %s" % ident)


def git_pull():
    """Updates the source with a git pull."""
    print 'Updating source from branch %s...' % builder.config.BRANCH
    os.chdir(builder.config.DISTRIB_DIR)
    run_git("git checkout " + builder.config.BRANCH)
    run_git("git pull")
    
    
def git_tag(tag):
    """Tags the source with a new tag."""
    print 'Tagging with %s...' % tag
    os.chdir(builder.config.DISTRIB_DIR)
    run_git("git tag %s" % tag, ignoreResult=True)
    run_git("git push --tags")



#-*- coding: UTF-8 -*-

"""
PyAssimp

This is the main-module of PyAssimp.
"""

import sys
if sys.version_info < (2,6):
    raise 'pyassimp: need python 2.6 or newer'


import ctypes
import os
import numpy

import logging; logger = logging.getLogger("pyassimp")

# Attach a default, null handler, to the logger.
# applications can easily get log messages from pyassimp
# by calling for instance
# >>> logging.basicConfig(level=logging.DEBUG)
# before importing pyassimp
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
h = NullHandler()
logger.addHandler(h)

from . import structs
from .errors import AssimpError
from . import helper


assimp_structs_as_tuple = (
        structs.Matrix4x4, 
        structs.Matrix3x3, 
        structs.Vector2D, 
        structs.Vector3D, 
        structs.Color3D, 
        structs.Color4D, 
        structs.Quaternion, 
        structs.Plane, 
        structs.Texel)

def make_tuple(ai_obj, type = None):
    res = None

    if isinstance(ai_obj, structs.Matrix4x4):
        res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_]).reshape((4,4))
        #import pdb;pdb.set_trace()
    elif isinstance(ai_obj, structs.Matrix3x3):
        res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_]).reshape((3,3))
    else:
        res = numpy.array([getattr(ai_obj, e[0]) for e in ai_obj._fields_])

    return res

# It is faster and more correct to have an init function for each assimp class
def _init_face(aiFace):
    aiFace.indices = [aiFace.mIndices[i] for i in range(aiFace.mNumIndices)]
    
assimp_struct_inits = \
    { structs.Face : _init_face }
    
def call_init(obj, caller = None):
    if helper.hasattr_silent(obj,'contents'): #pointer
        _init(obj.contents, obj, caller)
    else:
        _init(obj,parent=caller)

def _is_init_type(obj):
    if helper.hasattr_silent(obj,'contents'): #pointer
        return _is_init_type(obj[0])
    # null-pointer case that arises when we reach a mesh attribute
    # like mBitangents which use mNumVertices rather than mNumBitangents
    # so it breaks the 'is iterable' check.
    # Basically:
    # FIXME!
    elif not bool(obj): 
        return False
    tname = obj.__class__.__name__
    return not (tname[:2] == 'c_' or tname == 'Structure' \
            or tname == 'POINTER') and not isinstance(obj,int)
                    
def _init(self, target = None, parent = None):
    """
    Custom initialize() for C structs, adds safely accessable member functionality.

    :param target: set the object which receive the added methods. Useful when manipulating
    pointers, to skip the intermediate 'contents' deferencing.
    """
    if not target:
        target = self
    
    dirself = dir(self) 
    for m in dirself:

        if m.startswith("_"):
            continue

        if m.startswith('mNum'):
            if 'm' + m[4:] in dirself:
                continue # will be processed later on
            else:
                name = m[1:].lower()

                obj = getattr(self, m)
                setattr(target, name, obj)
                continue

        if m == 'mName':
            obj = self.mName
            target.name = str(obj.data.decode("utf-8"))
            target.__class__.__repr__ = lambda x: str(x.__class__) + "(" + x.name + ")"
            target.__class__.__str__ = lambda x: x.name
            continue
            
        name = m[1:].lower()

        obj = getattr(self, m)

        # Create tuples
        if isinstance(obj, assimp_structs_as_tuple):
            setattr(target, name, make_tuple(obj))
            logger.debug(str(self) + ": Added array " + str(getattr(target, name)) +  " as self." + name.lower())
            continue

        if m.startswith('m'):

            if name == "parent":
                setattr(target, name, parent)
                logger.debug("Added a parent as self." + name)
                continue

            if helper.hasattr_silent(self, 'mNum' + m[1:]):

                length =  getattr(self, 'mNum' + m[1:])
    
                # -> special case: properties are
                # stored as a dict.
                if m == 'mProperties':
                    setattr(target, name, _get_properties(obj, length))
                    continue


                if not length: # empty!
                    setattr(target, name, [])
                    logger.debug(str(self) + ": " + name + " is an empty list.")
                    continue
                

                try:
                    if obj._type_ in assimp_structs_as_tuple:
                        setattr(target, name, numpy.array([make_tuple(obj[i]) for i in range(length)], dtype=numpy.float32))

                        logger.debug(str(self) + ": Added an array of numpy arrays (type "+ str(type(obj)) + ") as self." + name)

                    else:
                        setattr(target, name, [obj[i] for i in range(length)]) #TODO: maybe not necessary to recreate an array?

                        logger.debug(str(self) + ": Added list of " + str(obj) + " " + name + " as self." + name + " (type: " + str(type(obj)) + ")")

                        # initialize array elements
                        try:
                            init = assimp_struct_inits[type(obj[0])]
                        except KeyError:
                            if _is_init_type(obj[0]):
                                for e in getattr(target, name):
                                    call_init(e, target)
                        else:
                            for e in getattr(target, name):
                                init(e)


                except IndexError:
                    logger.error("in " + str(self) +" : mismatch between mNum" + name + " and the actual amount of data in m" + name + ". This may be due to version mismatch between libassimp and pyassimp. Quitting now.")
                    sys.exit(1)

                except ValueError as e:
                    
                    logger.error("In " + str(self) +  "->" + name + ": " + str(e) + ". Quitting now.")
                    if "setting an array element with a sequence" in str(e):
                        logger.error("Note that pyassimp does not currently "
                                     "support meshes with mixed triangles "
                                     "and quads. Try to load your mesh with"
                                     " a post-processing to triangulate your"
                                     " faces.")
                    sys.exit(1)


            else: # starts with 'm' but not iterable

                setattr(target, name, obj)
                logger.debug("Added " + name + " as self." + name + " (type: " + str(type(obj)) + ")")
        
                if _is_init_type(obj):
                    call_init(obj, target)




    if isinstance(self, structs.Mesh):
        _finalize_mesh(self, target)

    if isinstance(self, structs.Texture):
        _finalize_texture(self, target)


    return self

class AssimpLib(object):
    """
    Assimp-Singleton
    """
    load, release, dll = helper.search_library()

#the loader as singleton
_assimp_lib = AssimpLib()

def pythonize_assimp(type, obj, scene):
    """ This method modify the Assimp data structures
    to make them easier to work with in Python.

    Supported operations:
     - MESH: replace a list of mesh IDs by reference to these meshes
     - ADDTRANSFORMATION: add a reference to an object's transformation taken from their associated node.

    :param type: the type of modification to operate (cf above)
    :param obj: the input object to modify
    :param scene: a reference to the whole scene
    """

    if type == "MESH":
        meshes = []
        for i in obj:
            meshes.append(scene.meshes[i])
        return meshes

    if type == "ADDTRANSFORMATION":

        def getnode(node, name):
            if node.name == name: return node
            for child in node.children:
                n = getnode(child, name)
                if n: return n

        node = getnode(scene.rootnode, obj.name)
        if not node:
            raise AssimpError("Object " + str(obj) + " has no associated node!")
        setattr(obj, "transformation", node.transformation)


def recur_pythonize(node, scene):
    """ Recursively call pythonize_assimp on
    nodes tree to apply several post-processing to
    pythonize the assimp datastructures.
    """

    node.meshes = pythonize_assimp("MESH", node.meshes, scene)

    for mesh in node.meshes:
        mesh.material = scene.materials[mesh.materialindex]

    for cam in scene.cameras:
        pythonize_assimp("ADDTRANSFORMATION", cam, scene)

    #for light in scene.lights:
    #    pythonize_assimp("ADDTRANSFORMATION", light, scene)

    for c in node.children:
        recur_pythonize(c, scene)


def load(filename, processing=0):
    """
    Loads the model with some specific processing parameters.

    filename - file to load model from
    processing - processing parameters

    result Scene-object with model-data

    throws AssimpError - could not open file
    """
    #read pure data
    #from ctypes import c_char_p, c_uint
    #model = _assimp_lib.load(c_char_p(filename), c_uint(processing))
    model = _assimp_lib.load(filename.encode("ascii"), processing)
    if not model:
        #Uhhh, something went wrong!
        raise AssimpError("could not import file: %s" % filename)

    scene = _init(model.contents)

    recur_pythonize(scene.rootnode, scene)

    return scene

def release(scene):
    from ctypes import pointer
    _assimp_lib.release(pointer(scene))

def _finalize_texture(tex, target):
    setattr(target, "achformathint", tex.achFormatHint)
    data = numpy.array([make_tuple(getattr(tex, "pcData")[i]) for i in range(tex.mWidth * tex.mHeight)])
    setattr(target, "data", data)

def _finalize_mesh(mesh, target):
    """ Building of meshes is a bit specific.

    We override here the various datasets that can
    not be process as regular fields.

    For instance, the length of the normals array is
    mNumVertices (no mNumNormals is available)
    """
    nb_vertices = getattr(mesh, "mNumVertices")

    def fill(name):
        mAttr = getattr(mesh, name)
        if mAttr:
            data = numpy.array([make_tuple(getattr(mesh, name)[i]) for i in range(nb_vertices)], dtype=numpy.float32)
            setattr(target, name[1:].lower(), data)
        else:
            setattr(target, name[1:].lower(), numpy.array([], dtype="float32"))

    def fillarray(name):
        mAttr = getattr(mesh, name)

        data = []
        for index, mSubAttr in enumerate(mAttr):
            if mSubAttr:
                data.append([make_tuple(getattr(mesh, name)[index][i]) for i in range(nb_vertices)])

        setattr(target, name[1:].lower(), numpy.array(data, dtype=numpy.float32))

    fill("mNormals")
    fill("mTangents")
    fill("mBitangents")

    fillarray("mColors")
    fillarray("mTextureCoords")
    
    # prepare faces
    faces = numpy.array([f.indices for f in target.faces], dtype=numpy.int32)
    setattr(target, 'faces', faces)


class PropertyGetter(dict):
    def __getitem__(self, key):
        semantic = 0
        if isinstance(key, tuple):
            key, semantic = key

        return dict.__getitem__(self, (key, semantic))

    def keys(self):
        for k in dict.keys(self):
            yield k[0]

    def __iter__(self):
        return self.keys()

    def items(self):
        for k, v in dict.items(self):
            yield k[0], v


def _get_properties(properties, length): 
    """
    Convenience Function to get the material properties as a dict
    and values in a python format.
    """
    result = {}
    #read all properties
    for p in [properties[i] for i in range(length)]:
        #the name
        p = p.contents
        key = (str(p.mKey.data.decode("utf-8")).split('.')[1], p.mSemantic)

        #the data
        from ctypes import POINTER, cast, c_int, c_float, sizeof
        if p.mType == 1:
            arr = cast(p.mData, POINTER(c_float * int(p.mDataLength/sizeof(c_float)) )).contents
            value = [x for x in arr]
        elif p.mType == 3: #string can't be an array
            value = cast(p.mData, POINTER(structs.MaterialPropertyString)).contents.data.decode("utf-8")
        elif p.mType == 4:
            arr = cast(p.mData, POINTER(c_int * int(p.mDataLength/sizeof(c_int)) )).contents
            value = [x for x in arr]
        else:
            value = p.mData[:p.mDataLength]

        if len(value) == 1:
            [value] = value

        result[key] = value

    return PropertyGetter(result)

def decompose_matrix(matrix):
    if not isinstance(matrix, structs.Matrix4x4):
        raise AssimpError("pyassimp.decompose_matrix failed: Not a Matrix4x4!")
    
    scaling = structs.Vector3D()
    rotation = structs.Quaternion()
    position = structs.Vector3D()
    
    from ctypes import byref, pointer
    _assimp_lib.dll.aiDecomposeMatrix(pointer(matrix), byref(scaling), byref(rotation), byref(position))
    return scaling._init(), rotation._init(), position._init()



# $Id$
# Build script for Mac OS X
# Now uses py2app instead of BundleBuilder

from distutils.core import setup
import py2app

ver = file('VERSION').read().strip()

name = 'Doomsday Engine'
iconfile = 'snowberry.icns'
verstr = ver + '-macosx'

includes = ['webbrowser']

resources = [
    'build/addons',
    'build/conf',
    'build/graphics',
    'build/lang',
    'build/plugins',
    'build/profiles'
]

plist = dict(
    CFBundleIconFile           = iconfile,
    CFBundleName               = name,
    CFBundleShortVersionString = ver,
    CFBundleGetInfoString      = verstr,
    CFBundleExecutable         = name,
    CFBundleIdentifier         = 'net.dengine.snowberry'
)

opts = dict(py2app = dict(
    compressed = 1,
    iconfile = iconfile,
    plist = plist,
    includes = includes,
    optimize = 2
    ))

setup( app = ['snowberry.py'], data_files = resources, options = opts )



#!/usr/bin/env python2.7
# clang/gcc build log cleaner

import sys, re, string

def prefix_white(ln):
    pfx = ''
    for c in ln:
        if c in string.whitespace:
            pfx += c * 2
        else:
            break
    return pfx

lines   = file(sys.argv[1]).readlines()
erlines = file(sys.argv[2]).readlines()

print "Original log output contains %i+%i lines.\n" % (len(lines), len(erlines))

components = {}
qmakeOut = []

class Module:
    def __init__(self):
        self.name = ''
        self.location = ''
        self.cached = False

clangMode = False

for ln in lines:
    ln = ln.rstrip()

    # What kind of a line do we have?
    found = re.search('[a-z]+\.pro', ln)
    if found:
        print prefix_white(ln) + 'qmake ' + found.group(0)   
        continue

    #found = re.search('Entering directory.*/(deng/.*)\'', ln)
    #if found:
    #    print '\nBuilding in ' + found.group(1) + '...\n'
    #    continue

    if 'clang' in ln: clangMode = True
        
    found = re.search('.*(clang|gcc|g\+\+).*/doomsday/(.*/)([A-Za-z_0-9]+\.c[p]*)', ln)
    if found:
        mod = Module()
        mod.name = found.group(3)
        mod.location = found.group(2)
        mod.cached = ln.startswith('ccache')
        #print mod.name, mod.location, mod.cached
        if mod.location not in components:
            components[mod.location] = {}            
        components[mod.location][mod.name] = mod
        continue
        
class Issue:
    def __init__(self):
        self.name = ''
        self.location = ''
        self.kind = 'other'
        self.msg = []
        self.lineNumber = 0
        
# Look for errors.
issues = {}
current = None
for ln in erlines:
    ln = ln.rstrip()
    
    if current and clangMode:
        if re.match('^[~^ ]+$', ln): 
            current = None
        else:
            current.msg.append(ln.strip())
        continue
    
    found = re.search('Project (MESSAGE|ERROR): (.*)', ln)
    if found:
        qmakeOut.append(found.group(2))
        continue    

    found = re.search('^([a-z0-9_/\.\-]+/[a-z_0-9]+\.[ch][p]*):[0-9]+', ln, re.IGNORECASE)
    if found:
        grp = re.search('/doomsday/(.*/)([a-z_0-9]+\.[ch][p]*)', found.group(1), re.IGNORECASE)
        if grp:
            kind = 'other'
            if ' warning:' in ln:
                kind = 'warning'
            elif ' error:' in ln:
                kind = 'error'
                
            msg = ''
            num = 0
            found = re.search('[chp]:([0-9]+):.*(warning|error): (.*)', ln)
            if found:
                num = int(found.group(1))
                msg = found.group(3)
                
            current = Issue()
            current.name = grp.group(2)
            current.location = grp.group(1)
            current.kind = kind
            current.lineNumber = num
            current.msg.append(msg.strip())
            
            if current.name not in issues: issues[current.name] = []
            issues[current.name].append(current)
            
            #print kind, grp.group(1), grp.group(2), msg
                
        

##############################################################################
# Produce output
##############################################################################
print
print string.join(qmakeOut, '\n')

# Group by component.
for compName in sorted(components):
    print '\n%-35s [%3i files]' % (compName, len(components[compName]))
    compIssues = []
    for modName in sorted(components[compName]):
        mod = components[compName][modName]
        # Any issues for this?
        status = ''
        warns = 0
        errs = 0
        if mod.name in issues:
            for issue in issues[mod.name]:
                if issue.location != compName: continue
                compIssues.append(issue)
                if issue.kind == 'error': errs += 1
                elif issue.kind == 'warning': warns += 1
        if errs > 0:
            status += '%i errors' % errs
        if warns > 0:
            if errs > 0: status += ' and '
            status += '%i warnings' % warns
        if len(status) == 0: #continue # Nothing to see here.
            status = 'ok'
        # Flags.
        flags = ''
        if mod.cached: flags = 'c'
        if len(flags) > 0: flags = '[%s]' % flags
        # Print it out.
        print ' %3s %-31s : %s' % (flags, mod.name, status)
        
    # Any issues for this component?
    if len(compIssues) > 0:
        for issue in compIssues:
            print '\n     %7s @ %s:%i: %s' % (issue.kind.upper(), issue.name, issue.lineNumber, string.join(issue.msg, '\n         '))

#for name in issues:
#    uniq = []
#    for issue in issues[name]:
#        combined = (name, issue.kind, issue.lineNumber, issue.msg)
#        if combined not in uniq: uniq.append(combined)
#    for name, kind, num, msg in uniq:
#        print name, kind, num, msg



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file wizard.py Configuration Wizard

import ui, paths, events, language, sb.util.dialog
from sb.util.dialog import WizardPage, WizardDialog
import sb.widget.button as wb
import sb.widget.text as wt
import sb.widget.list
import sb.confdb as st
import sb.profdb as pr
import sb.aodb as ao


HAS_BEEN_RUN = 'setup-wizard-shown' # see profdb.reset()


def init():
    # Register a listener for detecting the completion of Snowberry
    # startup.
    events.addNotifyListener(handleNotify, ['init-done', 'wizard-selected'])

    # Listen for wizard commands.
    events.addCommandListener(handleCommand, ['run-setup-wizard'])
    
    # Commands for the popup menu.
    ui.addMenuCommand(ui.MENU_TOOLS, 'run-setup-wizard', group=ui.MENU_GROUP_APP)    


def handleNotify(event):
    """Handle notifications."""

    if event.hasId('init-done'):
        # If this is the first launch ever, we'll display the wizard.
        if not pr.getDefaults().getValue(HAS_BEEN_RUN):
            runWizard()
        

def handleCommand(event):
    """Handle commands."""
    
    if event.hasId('run-setup-wizard'):
        runWizard()


class ProfileWizardPage (WizardPage):
    """A profile page is skipped if it's not checked in the game list."""

    def __init__(self, parentWizard, identifier, gameList):
        WizardPage.__init__(self, parentWizard, identifier)
        self.gameList = gameList
        self.iwadText = None

    def setGameList(self, list):
        self.gameList = list

    def update(self):
        if self.iwadText:
            self.iwadText.getFromProfile(pr.getActive())
            self.iwadText.select()

    def getAdjacent(self, theNext):
        next = self
        # Loop until a good page is found.
        while True:
            # Which direction are we going?
            if theNext:
                next = WizardPage.GetNext(next)
            else:
                next = WizardPage.GetPrev(next)

            # Should we skip or stop?
            if (next != None and self.gameList != None and
                next.getId() in self.gameList.getItems()):
                if next.getId() in self.gameList.getSelectedItems():
                    # Activate the correct profile.
                    break
            else:
                # Not a profile page.
                break
        return next

    def GetNext(self):
        return self.getAdjacent(True)
    
    def GetPrev(self):
        return self.getAdjacent(False)


def runWizard():
    """Run the wizard dialog."""

    # Make sure the help panel isn't updated during the wizard.
    #events.send(events.Command('freeze'))

    suggested = {
        'doom1': 'DOOM.WAD',
        'doom1-share': 'DOOM1.WAD',
        'doom1-ultimate': 'DOOM.WAD',
        'doom2': 'DOOM2.WAD',
        'doom2-tnt': 'TNT.WAD',
        'doom2-plut': 'PLUTONIA.WAD',
        'heretic-ext': 'HERETIC.WAD',
        'heretic-share': 'HERETIC.WAD',
        'heretic': 'HERETIC.WAD',
        'hexen': 'HEXEN.WAD',
        'hexen-demo': 'HEXEN.WAD',
        'hexen-dk': 'HEXEN.WAD',
        'hacx': 'HACX.WAD',
        'chex': 'CHEX.WAD'
    }
    
    events.mute()
    
    # Make the Defaults profile active.
    pr.setActive(pr.getDefaults())
    
    wiz = WizardDialog(language.translate('setup-wizard'),
                       paths.findBitmap('wizard'))

    # Language selection page.
    langPage = WizardPage(wiz, 'wizard-language')
    area = langPage.getArea()
    area.createText('wizard-language-explanation', maxLineLength=65).resizeToBestSize()
    sar, languageCheckBox = area.createSetting(st.getSetting('language'))
    languageCheckBox.getFromProfile(pr.getActive())

    # Game selection.
    gamePage = ProfileWizardPage(wiz, 'wizard-games', None)
    gamePage.follows(langPage)
    area = gamePage.getArea()
    area.createText('wizard-select-games', maxLineLength=65).resizeToBestSize()
    area.setBorderDirs(ui.BORDER_NOT_BOTTOM)
    games = area.createList('', style=sb.widget.list.List.STYLE_CHECKBOX)
    area.setBorderDirs(ui.BORDER_NOT_TOP)
    games.setMinSize(50, 180)
    gamePage.setGameList(games)

    def allGames():
        # Check the entire list.
        for item in games.getItems():
            games.checkItem(item, True)

    def clearGames():
        # Uncheck the entire list.
        for item in games.getItems():
            games.checkItem(item, False)

    
    controls = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=2)
    controls.setWeight(0)
    button = controls.createButton('wizard-games-all', wb.Button.STYLE_MINI)
    button.addReaction(allGames)
    button.resizeToBestSize()
    button = controls.createButton('wizard-games-clear', wb.Button.STYLE_MINI)
    button.addReaction(clearGames)
    button.resizeToBestSize()

    # The pages will be linked together.
    previousPage = gamePage

    deathKingsWad = None

    # We'll do this dynamically.
    checkedProfiles = ['doom1', 'doom2', 'heretic', 'hexen']
    # Only display the system profiles in the wizard (not any user
    # profiles).
    profiles = pr.getProfiles(lambda p: p.isSystemProfile())
    for p in profiles:
        if p is not pr.getDefaults():
            games.addItem(p.getId(), p.getId() in checkedProfiles)

            # Create a page for the profile.
            page = ProfileWizardPage(wiz, p.getId(), games)

            # Link the pages together.
            page.follows(previousPage)
            previousPage = page
            
            area = page.getArea()

            area.createText('wizard-locate-iwad', maxLineLength=65).resizeToBestSize()

            # The suggestion.
            if suggested.has_key(p.getId()):
                sugArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=2)
                sugArea.setExpanding(False)
                sugArea.setWeight(1)
                sugArea.createText('wizard-suggested-iwad', ':', align=wt.Text.RIGHT)
                sugArea.setWeight(2)
                sug = sugArea.createText('')
                sug.setText(suggested[p.getId()])

            sar, page.iwadText = area.createSetting(st.getSetting('iwad'))

            if p.getId() == 'hexen-dk':
                area.setBorder(12, ui.BORDER_ALL)
                area.createLine()
                area.setBorder(6, ui.BORDER_ALL)
                area.createText('deathkings-selection-title').setHeadingStyle()
                
                # Death Kings is an extension to Hexen.  It uses the
                # same Hexen IWAD, but also an addon IWAD.
                area.createText('wizard-locate-iwad-deathkings', maxLineLength=65).resizeToBestSize()

                entry = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=4)
                entry.setExpanding(False)
                entry.setWeight(1)
                entry.setBorderDirs(ui.BORDER_NOT_LEFT)
                deathKingsWad = entry.createTextField('')
                entry.setWeight(0)
                entry.setBorderDirs(ui.BORDER_TOP_BOTTOM)
                browseButton = entry.createButton('browse-button', wb.Button.STYLE_MINI)

                def browseDeathKings():
                    # Open a file browser.
                    selection = sb.util.dialog.chooseFile('deathkings-selection-title',
                                                          '', True,
                                                          [('file-type-wad', 'wad')])
                
                    if len(selection) > 0:
                        deathKingsWad.setText(selection)

                browseButton.addReaction(browseDeathKings)

	# Addon paths.
	pathsPage = ProfileWizardPage(wiz, 'wizard-addon-paths', games)
	pathsPage.follows(previousPage)
	area = pathsPage.getArea()
	area.createText('wizard-addon-paths-explanation',
                    maxLineLength=65).resizeToBestSize()

    area.setBorderDirs(ui.BORDER_NOT_BOTTOM)
    pathList = area.createList('addon-paths-list')
    pathList.setMinSize(100, 120)

    # Insert the current custom paths into the list.
    for p in paths.getAddonPaths():
        pathList.addItem(p)

    def addAddonPath():
        selection = sb.util.dialog.chooseFolder('addon-paths-add-prompt', '')
        if selection:
            pathList.addItem(selection)
            pathList.selectItem(selection)

    def removeAddonPath():
        selection = pathList.getSelectedItem()
        if selection:
            pathList.removeItem(selection)

    area.setWeight(0)
    area.setBorderDirs(ui.BORDER_NOT_TOP)
    commands = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=2)
    commands.setWeight(0)

    # Button for adding new paths.
    button = commands.createButton('new-addon-path', wb.Button.STYLE_MINI)
    button.addReaction(addAddonPath)

    # Button for removing a path.
    button = commands.createButton('delete-addon-path', wb.Button.STYLE_MINI)
    button.addReaction(removeAddonPath)

    # Launch options.
    quitPage = WizardPage(wiz, 'wizard-launching') 
    quitPage.follows(pathsPage)
    area = quitPage.getArea()
    area.createText('wizard-launching-explanation').resizeToBestSize()
    sar, quitCheckBox = area.createSetting(st.getSetting('quit-on-launch'))
    quitCheckBox.getFromProfile(pr.getActive())

    # List of unusable profiles, due to a missing IWAD.
    unusableProfiles = []
    
    # When the page changes in the wizard, change the active profile.
    def changeActiveProfile(page):
        prof = pr.get(page.getId())
        if prof:
            pr.setActive(prof)
        else:
            pr.setActive(pr.getDefaults())
        # Update page with values from the current profile.
        page.update()
                       
    wiz.setPageReaction(changeActiveProfile)
    
    if wiz.run(langPage) == 'ok':
        events.unmute()
        
        # Show the profiles that now have an IWAD.
        for prof in profiles:
            if prof.getValue('iwad', False) != None:
                pr.show(prof.getId())
                if prof.getId() in unusableProfiles:
                    unusableProfiles.remove(prof.getId())
            else:
                pr.hide(prof.getId())
                if prof.getId() not in unusableProfiles and \
                   prof.getId() in games.getSelectedItems():
                    unusableProfiles.append(prof.getId())

        # Install the Death Kings WAD?
        if deathKingsWad:
            try:
                ident = ao.install(deathKingsWad.getText())

                # Attach the WAD as an addon.
                kingsProfile = pr.get('hexen-dk')
                kingsProfile.useAddon(ident)
            except:
                # TODO: Handle error.
                pass

        # Update custom addon paths.
        currentAddonPaths = paths.getAddonPaths()
        wasModified = False
        for path in pathList.getItems():
            if path not in currentAddonPaths:
                paths.addAddonPath(path)
                wasModified = True
        for path in currentAddonPaths:
            if path not in pathList.getItems():
                paths.removeAddonPath(path)
                wasModified = True

        if wasModified:
            # Load addons from new paths.
            ao.refresh()

        events.send(events.Notify('addon-paths-changed'))

        # The wizard will only be shown once automatically.
        pr.getDefaults().setValue(HAS_BEEN_RUN, 'yes', False)

    else:
        # Wizard was canceled.
        events.unmute()

    pr.refresh()

    # This'll destroy all the pages of the wizard as well.
    wiz.destroy()

    # Enable help panel updates again.
    #events.send(events.Command('unfreeze'))
    
    # Tell the user about unusable profiles.
    if len(unusableProfiles) > 0:
        dialog, area = sb.util.dialog.createButtonDialog(
            'wizard-unlaunchable-profiles',
            ['ok'], 'ok', resizable=False)
        # Compose a list of the unlaunchable profiles.
        profList = ''
        unusableProfiles.sort(lambda a, b: cmp(language.translate(a),
                                               language.translate(b)))
        for p in unusableProfiles:
            profList += "\n" + language.translate(p)
        msg = area.createText()
        msg.setText(language.translate('wizard-unlaunchable-profiles-listed') + "\n" + 
                    profList)
        msg.resizeToBestSize()
        dialog.run()
        



#!/usr/bin/env python2.7
# coding=utf-8
#
# Script for performing automated build events.
# http://dengine.net/dew/index.php?title=Automated_build_system
#
# The Build Pilot (pilot.py) is responsible for task distribution 
# and management.

import sys
import os
import subprocess
import shutil
import time
import string
import glob
import builder
import pickle
import zipfile
from builder.git import * 
from builder.utils import * 
    
    
def pull_from_branch():
    """Pulls commits from the repository."""
    git_pull()
    
    
def create_build_event():
    """Creates and tags a new build for with today's number."""
    print 'Creating a new build event.'
    git_pull()

    # Identifier/tag for today's build.
    todaysBuild = todays_build_tag()
    
    # Tag the source with the build identifier.
    git_tag(todaysBuild)
        
    # Prepare the build directory.
    ev = builder.Event(todaysBuild)
    ev.clean()
    
    # Save the version number and release type.
    import build_version
    build_version.find_version(quiet=True)
    print >> file(ev.file_path('version.txt'), 'wt'), \
        build_version.DOOMSDAY_VERSION_FULL
    print >> file(ev.file_path('releaseType.txt'), 'wt'), \
        build_version.DOOMSDAY_RELEASE_TYPE        

    update_changes()
    

def mac_meta_update():
    """Update the OS X installer metadata."""
    import build_version
    build_version.find_version(quiet=True)
    system_command("deng_generate_installer_xml.py " + build_version.DOOMSDAY_VERSION_FULL_PLAIN)
    

def todays_platform_release():
    """Build today's release for the current platform."""
    print "Building today's build."
    ev = builder.Event()

    git_pull()

    git_checkout(ev.tag() + builder.config.TAG_MODIFIER)
    
    # We'll copy the new files to the build dir.
    os.chdir(builder.config.DISTRIB_DIR)
    oldFiles = DirState('releases', subdirs=False)
    
    try:
        print 'platform_release.py...'
        run_python2("platform_release.py > %s 2> %s" % \
            ('buildlog.txt', 'builderrors.txt'))
    except Exception, x:
        print 'Error during platform_release:', x
            
    for n in DirState('releases', subdirs=False).list_new_files(oldFiles):
        # Copy any new files.
        remote_copy(os.path.join('releases', n), ev.file_path(n))

        if builder.config.APT_REPO_DIR:
            # Copy also to the appropriate apt directory.
            arch = 'i386'
            if '_amd64' in n: arch = 'amd64'
            remote_copy(os.path.join('releases', n),
                        os.path.join(builder.config.APT_REPO_DIR, 
                                     builder.config.APT_DIST + \
                                     '/main/binary-%s' % arch, n))
                                 
    # Also the build logs.
    if sys.platform[:3] != 'win':
        system_command('./logcleanup.py buildlog.txt builderrors.txt > buildsummary.txt')
        remote_copy('buildsummary.txt', ev.file_path('doomsday-out-%s.txt' % sys_id()))
    else:        
        remote_copy('buildlog.txt', ev.file_path('doomsday-out-%s.txt' % sys_id()))
        remote_copy('builderrors.txt', ev.file_path('doomsday-err-%s.txt' % sys_id()))

    if 'linux' in sys_id():
        remote_copy('dsfmod/fmod-out-%s.txt' % sys_id(), ev.file_path('fmod-out-%s.txt' % sys_id()))
        remote_copy('dsfmod/fmod-err-%s.txt' % sys_id(), ev.file_path('fmod-err-%s.txt' % sys_id()))
                                             
    git_checkout(builder.config.BRANCH)


def sign_packages():
    """Sign all packages in the latest build."""
    ev = builder.Event(latestAvailable=True)
    print "Signing build %i." % ev.number()    
    for fn in os.listdir(ev.path()):
        if fn.endswith('.exe') or fn.endswith('.dmg') or fn.endswith('.deb'):
            # Make a signature for this.
            os.system("gpg --output %s -ba %s" % (ev.file_path(fn) + '.sig', ev.file_path(fn)))


def publish_packages():
    """Publish all packages to SourceForge."""
    ev = builder.Event(latestAvailable=True)
    print "Publishing build %i." % ev.number()
    system_command('deng_copy_build_to_sourceforge.sh "%s"' % ev.path())
    

def find_previous_tag(toTag, version):
    """Finds the build tag preceding `toTag`.
    
    Arguments:
        toTag:   Build tag ("buildNNNN").
        version: The preceding build tag must be from this version, 
                 comparing only major and minor version components.
                 Set to None to omit comparisons based on version.                 
    Returns:
        Build tag ("buildNNNN"). 
        None, if there is no applicable previous build. 
    """
    builds = builder.events_by_time() # descending by timestamp
    
    print "Finding previous build for %s (version:%s)" % (toTag, version)
         
    # Disregard builds later than `toTag`.
    while len(builds) and builds[0][1].tag() != toTag:
        del builds[0]
    if len(builds): del builds[0] # == toTag        
    if len(builds) == 0:
        return None
    
    for timestamp, ev in builds:
        print ev.tag(), ev.version(), time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(ev.timestamp()))
        
        if version is None:
            # Anything goes.
            return ev.tag()
        else:
            requiredVer = version_split(version)
            eventVer = version_split(ev.version())
            if requiredVer[:2] == eventVer[:2]:
                return ev.tag()            
        
    # Nothing suitable found. Fall back to a more lax search.
    return find_previous_tag(toTag, None)


def update_changes(debChanges=False):
    """Generates the list of commits for the latest build."""

    git_pull()
    toTag = todays_build_tag()

    import build_version

    # Look up the relevant version.
    event = builder.Event(toTag)
    refVersion = event.version()

    # Let's find the previous event of this version.
    fromTag = find_previous_tag(toTag, refVersion)
    
    if fromTag is None or toTag is None:
        # Range not defined.
        return

    print 'Changes for range', fromTag, '..', toTag

    changes = builder.Changes(fromTag, toTag)

    if debChanges:
        # Only update the Debian changelog.
        changes.generate('deb')

        # Also update the doomsday-fmod changelog (just version number).
        os.chdir(os.path.join(builder.config.DISTRIB_DIR, 'dsfmod'))
        
        fmodVer = build_version.parse_header_for_version('../../doomsday/plugins/fmod/include/version.h')
        debVer = "%s.%s.%s-%s" % (fmodVer[0], fmodVer[1], fmodVer[2], todays_build_tag())
        print "Marking new FMOD version:", debVer
        msg = 'New release: Doomsday Engine build %i.' % builder.Event().number()
        os.system('rm -f debian/changelog && dch --check-dirname-level 0 --create --package doomsday-fmod -v %s "%s"' % (debVer, msg))
        os.system('dch --release ""')

    else:
        changes.generate('html')
        changes.generate('xml')
           
           
def update_debian_changelog():
    """Updates the Debian changelog at (distrib)/debian/changelog."""
    # Update debian changelog.
    update_changes(debChanges=True)
           

def build_source_package():
    """Builds the source tarball and a Debian source package."""
    git_pull()
    ev = builder.Event(latestAvailable=True)
    print "Creating source tarball for build %i." % ev.number()
    os.chdir(os.path.join(builder.config.DISTRIB_DIR))
    remkdir('srcwork')
    os.chdir('srcwork')
    pkgName = 'doomsday'
    if ev.release_type() == 'stable':
        print 'Stable packages will be prepared'
        system_command('deng_package_source.sh ' + ev.version_base())
        pkgName += '-stable'
    else:
        system_command('deng_package_source.sh build %i %s' % (ev.number(), 
                                                               ev.version_base()))
    for fn in os.listdir('.'):
        if fn[:9] == 'doomsday-' and fn[-7:] == '.tar.gz' and ev.version_base() in fn:
            remote_copy(fn, ev.file_path(fn))
            break

    # Check distribution.
    system_command("lsb_release -a | perl -n -e 'm/Codename:\s(.+)/ && print $1' > /tmp/distroname")
    distro = file('/tmp/distroname', 'rt').read()

    # Create a source Debian package and upload it to Launchpad.
    pkgVer = '%s.%i+%s' % (ev.version_base(), ev.number(), distro)
    pkgDir = pkgName + '-%s' % pkgVer

    print "Extracting", fn
    system_command('tar xzf %s' % fn)
    print "Renaming", fn[:-7], 'to', pkgDir + '.orig'
    os.rename(fn[:-7], pkgDir + '.orig')

    origName = pkgName + '_%s' % ev.version_base() + '.orig.tar.gz'
    print "Symlink to", origName 
    system_command('ln %s %s' % (fn, origName))

    print "Extracting", fn
    system_command('tar xzf %s' % fn)
    print "Renaming", fn[:-7], 'to', pkgDir
    os.rename(fn[:-7], pkgDir)
    os.chdir(pkgDir)
    system_command('echo "" | dh_make -s -c gpl2 --file ../%s' % fn)
    os.chdir('debian')
    for fn in os.listdir('.'):
        if fn[-3:].lower() == '.ex': os.remove(fn)
    os.remove('README.Debian')
    os.remove('README.source')
    
    def gen_changelog(src, dst, extraSub=''):
        system_command("sed 's/%s-build%i/%s/;%s' %s > %s" % (
                ev.version_base(), ev.number(), pkgVer, extraSub, src, dst))

    # Figure out the name of the distribution.
    dsub = ''
    if distro:
        dsub = 's/) precise;/) %s;/' % distro
    if pkgName != 'doomsday':
        if dsub: dsub += ';'
        dsub += 's/^doomsday /%s /' % pkgName

    gen_changelog('../../../debian/changelog', 'changelog', dsub)
    system_command("sed 's/${Arch}/i386 amd64/;s/${Package}/%s/' ../../../debian/control.template > control" % pkgName)
    system_command("sed 's/`..\/build_number.py --print`/%i/;s/..\/..\/doomsday/..\/doomsday/' ../../../debian/rules > rules" % ev.number())
    os.chdir('..')
    system_command('debuild -S')
    os.chdir('..')
    system_command('dput ppa:sjke/doomsday %s_%s_source.changes' % (pkgName, pkgVer))


def rebuild_apt_repository():
    """Rebuilds the Apt repository by running apt-ftparchive."""
    aptDir = builder.config.APT_REPO_DIR
    print 'Rebuilding the apt repository in %s...' % aptDir
    
    os.system("apt-ftparchive generate ~/Dropbox/APT/ftparchive.conf")
    os.system("apt-ftparchive -c %s release %s/%s > %s/%s/Release" % (builder.config.APT_CONF_FILE, aptDir, builder.config.APT_DIST, aptDir, builder.config.APT_DIST))
    os.chdir("%s/%s" % (aptDir, builder.config.APT_DIST))
    try:
        os.remove("Release.gpg")
    except OSError:
        # Never mind.
        pass
    os.system("gpg --output Release.gpg -ba Release")
    os.system("~/Dropbox/Scripts/mirror-tree.py %s %s" % (aptDir, os.path.join(builder.config.EVENT_DIR, 'apt')))


def write_index_html(tag):
    ev = builder.Event(tag)
    f = file(ev.file_path('index.html'), 'wt')
    print >> f, "<html>"
    print >> f, "<head><title>Build %i</title></head>" % ev.number()
    print >> f, "<body>"
    print >> f, "<h1>Build %i</h1>" % ev.number()
    print >> f, ev.html_description(False)
    print >> f, "</body>"
    print >> f, "</html>"


def update_feed():
    """Generate events.rss into the event directory."""
    
    feedName = os.path.join(builder.config.EVENT_DIR, "events.rss")
    print "Updating feed in %s..." % feedName
    
    out = file(feedName, 'wt')
    print >> out, '<?xml version="1.0" encoding="UTF-8"?>'
    print >> out, '<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom">'
    print >> out, '<channel>'
    
    print >> out, '<title>Doomsday Engine Builds</title>'
    print >> out, '<link>http://dengine.net/</link>'
    print >> out, '<atom:link href="%s/events.rss" rel="self" type="application/rss+xml" />' % builder.config.BUILD_URI
    print >> out, '<description>Automated binary builds of the Doomsday Engine.</description>'
    print >> out, '<language>en-us</language>'
    print >> out, '<webMaster>skyjake@users.sourceforge.net (Jaakko Keränen)</webMaster>'
    print >> out, '<lastBuildDate>%s</lastBuildDate>' % time.strftime(builder.config.RFC_TIME, 
        time.gmtime(builder.find_newest_event()['time']))
    print >> out, '<generator>autobuild.py</generator>'
    print >> out, '<ttl>180</ttl>' # 3 hours
    
    for timestamp, ev in builder.events_by_time():
        print >> out, '<item>'
        print >> out, '<title>Build %i</title>' % ev.number()
        print >> out, '<link>%s/%s/</link>' % ("http://dengine.net", ev.tag())
        print >> out, '<author>skyjake@users.sourceforge.net (skyjake)</author>'
        print >> out, '<pubDate>%s</pubDate>' % time.strftime(builder.config.RFC_TIME, time.gmtime(timestamp))
        print >> out, '<atom:summary>%s</atom:summary>' % ev.text_summary()
        print >> out, '<description>%s</description>' % ev.html_description()
        print >> out, '<guid isPermaLink="false">%s</guid>' % ev.tag()
        print >> out, '</item>'
        
        write_index_html(ev.tag())
    
    # Close.
    print >> out, '</channel>'
    print >> out, '</rss>'
    
    
def update_xml_feed():
    """Generate events.xml into the event directory."""
    
    feedName = os.path.join(builder.config.EVENT_DIR, "events.xml")
    print "Updating XML feed in %s..." % feedName
    
    out = file(feedName, 'wt')
    print >> out, '<?xml version="1.0" encoding="UTF-8"?>'
    print >> out, '<log>'
    for timestamp, ev in builder.events_by_time():
        print >> out, ev.xml_description()    
    print >> out, '</log>'
    

def purge_apt_repository(atLeastSeconds):
    for d in ['i386', 'amd64']:
        binDir = os.path.join(builder.config.APT_REPO_DIR, builder.config.APT_DIST + '/main/binary-') + d
        print 'Pruning binary apt directory', binDir
        # Find the old files.
        for fn in os.listdir(binDir):
            if fn[-4:] != '.deb': continue
            debPath = os.path.join(binDir, fn)
            ct = os.stat(debPath).st_ctime
            if time.time() - ct >= atLeastSeconds:
                print 'Deleting', debPath
                os.remove(debPath)


def purge_obsolete():
    """Purge old builds from the event directory (old > 12 weeks)."""
    threshold = 3600 * 24 * 7 * 12

    # Also purge the apt repository if one has been specified.
    if builder.config.APT_REPO_DIR:
        purge_apt_repository(threshold)
    
    # Purge the old events.
    print 'Deleting build events older than 12 weeks...'
    for ev in builder.find_old_events(threshold):
        print ev.tag()
        shutil.rmtree(ev.path()) 
        
    print 'Purge done.'


def dir_cleanup():
    """Purges empty build directories from the event directory."""
    print 'Event directory cleanup starting...'
    for emptyEventPath in builder.find_empty_events():
        print 'Deleting', emptyEventPath
        os.rmdir(emptyEventPath)
    print 'Cleanup done.'


def system_command(cmd):
    result = subprocess.call(cmd, shell=True)
    if result != 0:
        raise Exception("Error from " + cmd)


def generate_apidoc():
    """Run Doxygen to generate all API documentation."""
    git_pull()

    print >> sys.stderr, "\nSDK docs for Qt Creator..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday'))    
    system_command('doxygen sdk-qch.doxy >/dev/null 2>doxyissues-qch.txt')
    system_command('wc -l doxyissues-qch.txt')

    print >> sys.stderr, "\nPublic API 2.0 docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/libcore'))    
    system_command('doxygen api2.doxy >/dev/null 2>../doxyissues-api2.txt')
    system_command('wc -l ../doxyissues-api2.txt')
    
    print >> sys.stderr, "\nPublic API docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/client'))    
    system_command('doxygen api.doxy >/dev/null 2>../doxyissues-api.txt')
    system_command('wc -l ../doxyissues-api.txt')

    print >> sys.stderr, "\nlibshell docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/libshell'))
    system_command('doxygen shell.doxy >/dev/null 2>../doxyissues-shell.txt')
    system_command('wc -l ../doxyissues-shell.txt')

    print >> sys.stderr, "\nInternal Win32 docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/client'))
    system_command('doxygen client-win32.doxy >/dev/null 2>../doxyissues-win32.txt')
    system_command('wc -l ../doxyissues-win32.txt')

    print >> sys.stderr, "\nInternal Mac/Unix docs..."
    system_command('doxygen client-mac.doxy >/dev/null 2>../doxyissues-mac.txt')        
    system_command('wc -l ../doxyissues-mac.txt')

    print >> sys.stderr, "\nDoom plugin docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/plugins/doom'))
    system_command('doxygen doom.doxy >/dev/null 2>../../doxyissues-doom.txt')
    system_command('wc -l ../../doxyissues-doom.txt')

    print >> sys.stderr, "\nHeretic plugin docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/plugins/heretic'))
    system_command('doxygen heretic.doxy >/dev/null 2>../../doxyissues-heretic.txt')
    system_command('wc -l ../../doxyissues-heretic.txt')

    print >> sys.stderr, "\nHexen plugin docs..."
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/plugins/hexen'))
    system_command('doxygen hexen.doxy >/dev/null 2>../../doxyissues-hexen.txt')
    system_command('wc -l ../../doxyissues-hexen.txt')


def generate_readme():
    """Run Amethyst to generate readme documentation."""
    git_pull()
    os.chdir(os.path.join(builder.config.DISTRIB_DIR, '../doomsday/doc/output'))
    system_command('make clean all')
    
    
def generate_wiki():
    """Automatically generate wiki pages."""
    git_pull()
    sys.path += ['/Users/jaakko/Dropbox/Scripts']
    import dew
    dew.login()
    # Today's event data.
    ev = builder.Event(latestAvailable=True)
    if ev.release_type() == 'stable':
        dew.submitPage('Latest Doomsday release',
            '#REDIRECT [[Doomsday version %s]]' % ev.version())
    dew.logout()
    

def web_path():
    return os.path.join(builder.config.DISTRIB_DIR, '..', 'web')
    
    
def web_manifest_filename():
    return os.path.join(builder.config.DISTRIB_DIR, '..', '.web.manifest')
    

def web_save(state):
    pickle.dump(state, file(web_manifest_filename(), 'wb'), pickle.HIGHEST_PROTOCOL)
    
    
def web_init():
    print 'Initializing web update manifest.'
    web_save(DirState(web_path()))
    
    
def web_update():
    print 'Checking for web file changes...'
    git_pull()
    
    oldState = pickle.load(file(web_manifest_filename(), 'rb'))
    state = DirState(web_path())
    updated = state.list_new_files(oldState)
    removed = state.list_removed(oldState)
    
    # Save the updated state.
    web_save(state)
    
    # Is there anything to do?
    if not updated and not removed[0] and not removed[1]:
        print 'Everything up-to-date.'
        return
    
    # Compile the update package.
    print 'Updated:', updated
    print 'Removed:', removed
    arcFn = os.path.join(builder.config.DISTRIB_DIR, '..', 
        'web_update_%s.zip' % time.strftime('%Y%m%d-%H%M%S'))
    arc = zipfile.ZipFile(arcFn, 'w')
    for up in updated:
        arc.write(os.path.join(web_path(), up), 'updated/' + up)
    if removed[0] or removed[1]:
        tmpFn = '_removed.tmp'
        tmp = file(tmpFn, 'wt')
        for n in removed[0]:        
            print >> tmp, 'rm', n
        for d in removed[1]:
            print >> tmp, 'rmdir', d
        tmp.close()
        arc.write(tmpFn, 'removed.txt')
        os.remove(tmpFn)
    arc.close()
    
    # Deliver the update to the website.
    print 'Delivering', arcFn
    system_command('scp %s dengine@dengine.net:www/incoming/' % arcFn)
    
    # No need to keep a local copy.
    os.remove(arcFn)
    

def show_help():
    """Prints a description of each command."""
    for cmd in sorted_commands():
        if commands[cmd].__doc__:
            print "%-17s " % (cmd + ":") + commands[cmd].__doc__
        else:
            print cmd
    

def sorted_commands():
    sc = commands.keys()
    sc.sort()
    return sc


commands = {
    'pull': pull_from_branch,
    'create': create_build_event,
    'platform_release': todays_platform_release,
    'sign': sign_packages,
    'publish': publish_packages,
    'changes': update_changes,
    'debchanges': update_debian_changelog,
    'source': build_source_package,
    'apt': rebuild_apt_repository,
    'feed': update_feed,
    'xmlfeed': update_xml_feed,
    'purge': purge_obsolete,
    'cleanup': dir_cleanup,
    'apidoc': generate_apidoc,
    'readme': generate_readme,
    'wiki': generate_wiki,
    'web_init': web_init,
    'web_update': web_update,
    'mac_meta': mac_meta_update,
    'help': show_help
}

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print 'The arguments must be: (command) [args]'
        print 'Commands:', string.join(sorted_commands())
        print 'Arguments:'
        print '--branch   Branch to use (default: master)'
        print '--distrib  Doomsday distrib directory'
        print '--events   Event directory (builds are stored here in subdirs)'
        print '--apt      Apt repository'
        print '--tagmod   Additional suffix for build tag for platform_release'
        sys.exit(1)

    if sys.argv[1] not in commands:
        print 'Unknown command:', sys.argv[1]
        sys.exit(1)

    commands[sys.argv[1]]()



import os, sys

comps = ['libdoom', 'libheretic', 'libhexen']

def unify(mode):
    for fn in os.listdir('libdoom/'+mode):
        # Do the others have this?
        content = file('libdoom/%s/%s' % (mode, fn)).read()

        try:
            if content == file('libheretic/%s/%s' % (mode, fn)).read() and \
               content == file('libhexen/%s/%s' % (mode, fn)).read():
           
               file('libcommon/%s/%s' % (mode, fn), 'w').write(content)
               for com in comps:
                   os.remove('%s/%s/%s' % (com, mode, fn))        
               print 'unified', mode, fn
        except IOError:
            pass
        
unify('variable')
unify('command')



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#   Antti Kopponen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file profile.py Profile Class

import os, re, string, shutil
import events, paths, cfparser, language, logger
import sb.confdb as st
import aodb
import profdb


class Value:
    """Instances of the Value class are stored inside Profile objects.
    Each Value is essentially a tuple that pairs a setting identifier
    with a string value."""

    def __init__(self, settingId, value):
        self.settingId = settingId
        self.value = value

    def getId(self):
        return self.settingId

    def getValue(self):
        return self.value

    def setValue(self, newValue):
        self.value = newValue


class Profile:
    """A Profile is a collection of information about setting values
    and addons marked for loading."""

    def __init__(self, identifier):
        """Construct a new empty Profile object.  The profile will
        contain no values or attached addons.

        @param identifier The identifier of the profile.  This
        determines the file name of the profile file.  Identifiers are
        always lower case strings.
        """
        self.identifier = identifier.lower()
        self.name = ''
        self.values = []
        self.addons = []
        self.loadOrder = []
        self.story = ''
        self.hidden = False
        self.banner = 'banner-default'

    def duplicate(self, newIdentifier):
        """Create a duplicate of the profile.

        @param newIdentifier Identifier of the new Profile object.

        @return The newly created Profile object.
        """

        p = Profile(newIdentifier)
        p.name = self.name
        for v in self.values:
            p.values.append(v)
        for a in self.addons:
            p.addons.append(a)
        for o in self.loadOrder:
            p.loadOrder.append(o)
        p.story = self.story
        p.hidden = self.hidden
        p.banner = self.banner
        return p

    def getId(self):
        """Returns the identifier of the profile.  This is determined
        by the file name of the profile.
        """
        return self.identifier

    def getName(self):
        """Returns the name of the profile.  This is the name provided
        by the user.
        """
        if self is profdb.defaults:
            # The defaults profile has a translateable name.
            return language.translate('defaults-profile')
        
        return self.name

    def getLoadOrder(self):
        return self.loadOrder

    def setLoadOrder(self, order):
        """Set the load order for the profile.

        @param order An array of addon identifiers.
        """
        
        self.loadOrder = order

    def isHidden(self):
        return self.hidden

    def setHidden(self, hide):
        """Hidden profiles aren't listed in the profiles list."""
        self.hidden = hide

    def isSystemProfile(self):
        """Checks if this profile is a system profile.  System profiles can 
        only be hidden, not removed completely.

        @return True, if this is a system profile.
        """
        locations = [paths.getSystemPath(paths.PROFILES)] + \
                     paths.getBundlePaths(paths.PROFILES)
        for loc in locations:
            if os.path.exists(os.path.join(loc, self.getFileName())):
                return True
        return False

    def getFileName(self):
        """Return the file name this profile uses."""
        return self.identifier + ".prof"

    def setName(self, name):
        self.name = name

        # Define this in the english language so it will available for
        # translation.
        language.define('english', self.identifier, name)

    def getAllValues(self):
        """Returns a list of all the values defined in the profile.

        @return An array of Value objects.
        """
        return self.values

    def hasValue(self, id):
        "Checks if the specified setting Id has a value in this profile."
        for v in self.values:
            if v.getId() == id:
                return True
        return False

    def getValue(self, id, fallbackToDefaults=True):
        """Returns the Value for the given setting. Falls back to the
        defaults profile.

        @param id Setting identifier.

        @param fallbackToDefaults If True, the value is retrieved from
        the Defaults profile if it isn't defined in this one.
        """
        defaults = profdb.defaults

        # If this is a system setting, let's consult the system
        # configuration.
        if st.isDefined(id):
            return Value(id, st.getSystemString(id))

        # First look in the Values in this profile.
        for v in self.values:
            if v.getId() == id:
                return v
                
        if id == 'game-mode':
            # We can look for the game mode in the system profile.
            sysProfFileName = os.path.join(paths.getSystemPath(paths.PROFILES), self.getFileName())
            if os.path.exists(sysProfFileName):
                try:
                    p = cfparser.FileParser(sysProfFileName)
                    while True:
                        element = p.get()
                        if element.isKey() and element.getName() == 'game-mode':
                            return Value(element.getName(), element.getValue())
                except cfparser.OutOfElements:
                    # The file ended normally.
                    pass           

        # Fall back to the Defaults profile.
        if fallbackToDefaults and self is not defaults:
            return defaults.getValue(id)
        elif self is defaults:
            # See if the setting itself has a default for this.
            try:
                setting = st.getSetting(id)
                if setting.getDefault():
                    return Value(id, str(setting.getDefault()))
            except:
                pass
            # No Value for this setting could be found.
            return None
        else:
            # A normal profile with no value for this setting.
            return None

    def setValue(self, settingId, newValue, doNotify=True):
        """Set the value of the setting in this profile.

        @param settingId Identifier of the setting to set.

        @param newValue New value for the setting (as a string).

        @param doNotify If True, a notification event will be sent.
        """
        value = None

        if st.isDefined(settingId):
            # System settings are not saved into profiles.
            return

        # Change the value of an existing Value.
        for v in self.values:
            if v.getId() == settingId:
                value = v

        if value:
            value.setValue(newValue)
        else:
            value = Value(settingId, newValue)
            self.values.append(value)

        # A kludge for detecting a change of the language.
        if self is profdb.defaults and settingId == 'language':
            language.change(newValue)
                
        # Send a notification of this.
        if doNotify:
            events.send(events.ValueNotify(settingId, newValue, self))

    def removeValue(self, settingId):
        """Remove this value from the profile entirely.

        @param settingId Identifier of the setting.
        """
        # Kludge for detecting the change of the language.
        if self is profdb.defaults and settingId == 'language':
            language.change(None)

        for v in self.values:
            if v.getId() == settingId:
                self.values.remove(v)
                events.send(events.ValueNotify(settingId, None, self))

    def setStory(self, id):
        self.story = id

    def getStory(self):
        return self.story

    def getBanner(self):
        return self.banner

    def setBanner(self, banner):
        self.banner = banner

    def setAddons(self, addonIdentifiers):
        """Set the addons associated with the profile, without sending
        notifications.  This is used when the profile is being loaded.

        @param addonIdentifiers An array of identifiers.
        """
        self.addons = addonIdentifiers

    def addAddon(self, identifier):
        """Attaches an addon to the profile.  If this is the Defaults
        profile, the addon is detached from all the other profiles.
        This is necessary because of the exclusion rule (see
        getUsedAddons()).

        @param identifier Addon identifier.
        """
        if identifier not in self.addons:
            self.addons.append(identifier)
            # Also send a notification.
            events.send(events.AttachNotify(identifier, self, True))

            if self is profdb.defaults:
                # Detach from others.  The appropriate detach events
                # will be transmitted.
                for p in profdb.getProfiles():
                    if p is not self:
                        p.removeAddon(identifier)

    def removeAddon(self, identifier):
        if identifier in self.addons:
            self.addons.remove(identifier)
            # Also send a notification.
            events.send(events.AttachNotify(identifier, self, False))

    def useAddon(self, identifier):
        """Higher-level addon attaching.  Takes care of the proper
        handling of defaults."""

        if aodb.get(identifier).isOptIn(self):
            self.addAddon(identifier)
        else:
            self.removeAddon(identifier)

    def dontUseAddon(self, identifier):
        """Higher-level addon detaching.  Takes care of the proper
        handling of defaults."""
        
        if aodb.get(identifier).isOptIn(self):
            self.removeAddon(identifier)
        else:
            self.addAddon(identifier)

    def getComponents(self):
        """Match the components with the values of the settings to
        figure out which components are being used.

        @return An array with the identifiers of the used components.
        """
        # Get all the settings that are used to select components.
        compSettings = st.getComponentSettings()

        # Collect the identifiers of the components we use into this
        # array.
        used = []

        for s in compSettings:
            value = self.getValue(s)
            if not value:
                # This component is not selected at all.
                continue
            used.append(value.getValue())

        return used

    def getAddons(self):
        """Returns the list of addons attached to the profile."""
        return [a for a in self.addons if aodb.exists(a)]

    def getUsedAddons(self):
        """Returns the list of all addons that will be used with the
        profile.  This includes the addons that have been attached to
        the Defaults profile.  If a addon is attached to both Defaults
        and this profile, it won't be used (exclusion).

        @return An array of addon identifiers.
        """
        if self is profdb.getDefaults():
            usedAddons = self.getAddons()
        else:
            usedAddons = []

            # Decide which addons are actually used.  The ones that are
            # both in defaultAddons and profileAddons are NOT used.
            # ((P union D) - (P intersection D))
            for a in self.getAddons():
                usedAddons.append(a)

            for d in profdb.getDefaults().getAddons():
                if d in self.getAddons():
                    # This addon is in both the Defaults and this profile,
                    # don't use it.
                    usedAddons.remove(d)
                else:
                    usedAddons.append(d)

        # Any inversed addons that should be used?
        for addon in aodb.getAddons():
            if addon.isInversed():
                if addon.getId() not in usedAddons:
                    usedAddons.append(addon.getId())
                else:
                    usedAddons.remove(addon.getId())
            
        # Filter out the incompatible ones.
        return [a for a in usedAddons if aodb.get(a).isCompatibleWith(self)]

    def getFinalAddons(self):
        """The final addons of a profile are the ones that will be
        used when a game is launched.  The returned list is sorted
        based on the load order that applies to the profile.

        For the Defaults profile, the final addons include ALL the
        existing addons.

        @return A list of addon identifiers, in the actual load order.
        """
        if self is profdb.defaults:
            # Return all existing addons.
            finalAddons = map(lambda a: a.getId(), aodb.getAddons())

        else:
            usedAddons = self.getUsedAddons()

            finalAddons = []

            # Any boxes?
            for ident in usedAddons:
                addon = aodb.get(ident)

                if addon.getBox():
                    # If the box hasn't been selected by the user, forget it.
                    if addon.getBox().getId() not in usedAddons:
                        continue
                
                if addon.isBox():
                    # Also include the required parts of the addons.
                    finalAddons += addon.getRequiredParts()

                finalAddons.append(ident)

        # Apply load order.
        self.sortByLoadOrder(finalAddons)

        return finalAddons                

    def getKeywords(self):
        """The values in the profile provide keywords that can be
        matched against addon dependencies."""

        keys = []

        for setting in st.getSettings():
            # Must be either a toggle or a choice.
            if setting.getType() != 'toggle' and setting.getType() != 'choice':
                continue

            # Must have a value defined.
            v = self.getValue(setting.getId())
            if not v:
                continue

            # Active toggles define a keyword.
            if setting.getType() == 'toggle' and v.getValue() == 'yes':
                if v.getId() not in keys:
                    keys.append(v.getId())

            # Choices define a keyword.
            if setting.getType() == 'choice':
                if v.getValue() not in keys:
                    keys.append(v.getValue())

            # Other setting types don't create keywords.

        return keys

    def sortByLoadOrder(self, addons):
        """Sort a list of addon identifiers based on the load order
        that applies to this profile.  The order is the result of
        multiple factors:
        -# Each profile can define its own load order.
        -# The default profile is consulted is a certain addon's proper
        order is not defined.
        -# If the default profile does not define an order, the
        priority class of the addon is used.

        @param addons The list of addon identifiers.  Will be
        modified.
        """
        def orderCompare(a, b):
            """Compare two addon identifiers.

            @param a Addon identifier.

            @param b Addon identifier.

            @return -1, if a comes before b.  1, if b comes before a.
            """
            # First check if order of these two addons is defined in
            # this profile.
            order = self.getLoadOrder()
            if a in order and b in order:
                # Both of them are in this order.
                return cmp(order.index(a), order.index(b))

            # Check the Defaults profile.
            if self is not profdb.defaults:
                order = profdb.defaults.getLoadOrder()
                if a in order and b in order:
                    # Both of them are in this order.
                    return cmp(order.index(a), order.index(b))

            # Compare the priority classes, then.
            return cmp(aodb.get(a).getPriority(), aodb.get(b).getPriority())

        # Sort them.
        addons.sort(orderCompare)
